
#
# spyne - Copyright (C) Spyne contributors.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
#

"""The ``spyne.decorator`` module contains the the @srpc decorator and its
helper methods. The @srpc decorator is responsible for tagging methods as remote
procedure calls extracting method's input and output types.

It's possible to create custom decorators that wrap the @srpc decorator in order
to have a more elegant way of passing frequently-used parameter values. The @rpc
decorator is a simple example of this.
"""

import spyne.const.xml

from copy import copy
from inspect import isclass

from spyne import MethodDescriptor

# Empty means empty input, bare output. Doesn't say anything about response
# being empty
from spyne import LogicError
from spyne import BODY_STYLE_EMPTY
from spyne import BODY_STYLE_WRAPPED
from spyne import BODY_STYLE_BARE
from spyne import BODY_STYLE_OUT_BARE
from spyne import BODY_STYLE_EMPTY_OUT_BARE

from spyne.model import ModelBase, ComplexModel, ComplexModelBase
from spyne.model.complex import TypeInfo, recust_selfref, SelfReference

from spyne.const import add_request_suffix


def _produce_input_message(f, params, in_message_name, in_variable_names,
                       no_ctx, no_self, argnames, body_style_str, self_ref_cls,
                       in_wsdl_part_name):
    arg_start = 0
    if no_ctx is False:
        arg_start += 1
    if no_self is False:
        arg_start += 1

    if argnames is None:
        try:
            argcount = f.__code__.co_argcount
            argnames = f.__code__.co_varnames[arg_start:argcount]

        except AttributeError:
            raise TypeError(
                "It's not possible to instrospect builtins. You must pass a "
                "sequence of argument names as the '_args' argument to the "
                "rpc decorator to manually denote the arguments that this "
                "function accepts."
            )

        if no_self is False:
            params = [self_ref_cls.novalidate_freq()] + params
            argnames = ('self',) + argnames

        if len(params) != len(argnames):
            raise LogicError("%r function has %d argument(s) but its decorator "
                           "has %d." % (f.__name__, len(argnames), len(params)))

    else:
        argnames = copy(argnames)
        if len(params) != len(argnames):
            raise LogicError("%r function has %d argument(s) but the _args "
                            "argument has %d." % (
                                f.__name__, len(argnames), len(params)))

    in_params = TypeInfo()
    from spyne import SelfReference
    for k, v in zip(argnames, params):
        try:
            is_self_ref = issubclass(v, SelfReference)
        except TypeError:
            is_self_ref = False

        if is_self_ref:
            if no_self is False:
                raise LogicError("SelfReference can't be used in @rpc")
            v = recust_selfref(v, self_ref_cls)

        k = in_variable_names.get(k, k)
        in_params[k] = v

    ns = spyne.const.xml.DEFAULT_NS
    if in_message_name.startswith("{"):
        ns, _, in_message_name = in_message_name[1:].partition("}")

    message = None
    if body_style_str == 'bare':
        if len(in_params) > 1:
            # The soap Body elt contains 1 elt (called "body entry" in the soap
            # standard) per method call. If bare methods were allowed to have >1
            # argument, it would have to be serialized as multiple body entries,
            # which would violate the standard. It's easy to work around this
            # restriction by creating a ComplexModel that contains all the
            # required parameters.
            raise LogicError("body_style='bare' can handle at most one "
                                                           "function argument.")

        if len(in_params) == 0:
            message = ComplexModel.produce(type_name=in_message_name,
                                           namespace=ns, members=in_params)
        else:
            message, = in_params.values()
            message = message.customize(sub_name=in_message_name, sub_ns=ns)

            if issubclass(message, ComplexModelBase) and not message._type_info:
                raise LogicError("body_style='bare' does not allow empty "
                                                               "model as param")

            # there can't be multiple arguments here.
            if message.__type_name__ is ModelBase.Empty:
                message._fill_empty_type_name(ns, in_message_name,
                                                    "%s_arg0" % in_message_name)

    else:
        message = ComplexModel.produce(type_name=in_message_name,
                                       namespace=ns, members=in_params)
        message.__namespace__ = ns

    if in_wsdl_part_name:
        message = message.customize(wsdl_part_name=in_wsdl_part_name)

    return message


def _validate_body_style(kparams):
    _body_style = kparams.pop('_body_style', None)
    _soap_body_style = kparams.pop('_soap_body_style', None)

    allowed_body_styles = ('wrapped', 'bare', 'out_bare')
    if _body_style is None:
        _body_style = 'wrapped'

    elif not (_body_style in allowed_body_styles):
        raise ValueError("body_style must be one of %r" %
                                                         (allowed_body_styles,))

    elif _soap_body_style == 'document':
        _body_style = 'wrapped'

    elif _soap_body_style == 'rpc':
        _body_style = 'bare'

    elif _soap_body_style is None:
        pass

    else:
        raise ValueError("soap_body_style must be one of ('rpc', 'document')")

    assert _body_style in ('wrapped', 'bare', 'out_bare')

    return _body_style


def _produce_output_message(func_name, body_style_str, self_ref_cls,
                                                              no_self, kparams):
    """Generate an output message for "rpc"-style API methods.

    This message is a wrapper to the declared return type.
    """

    _returns = kparams.pop('_returns', None)

    try:
        is_self_ref = issubclass(_returns, SelfReference)
    except TypeError:
        is_self_ref = False

    if is_self_ref:
        if no_self is False:
            raise LogicError("SelfReference can't be used in @rpc")

        _returns = recust_selfref(_returns, self_ref_cls)

    _is_out_message_name_overridden = not ('_out_message_name' in kparams)
    _out_message_name = kparams.pop('_out_message_name', '%s%s' %
                                       (func_name, spyne.const.RESPONSE_SUFFIX))

    if no_self is False and \
               (body_style_str == 'wrapped' or _is_out_message_name_overridden):
        _out_message_name = '%s.%s' % \
                               (self_ref_cls.get_type_name(), _out_message_name)

    _out_wsdl_part_name = kparams.pop('_wsdl_part_name', None)

    out_params = TypeInfo()

    if _returns and body_style_str == 'wrapped':
        if isinstance(_returns, (list, tuple)):
            default_names = ['%s%s%d'% (func_name, spyne.const.RESULT_SUFFIX, i)
                                                  for i in range(len(_returns))]

            _out_variable_names = kparams.pop('_out_variable_names',
                                                                  default_names)

            assert (len(_returns) == len(_out_variable_names))

            var_pair = zip(_out_variable_names, _returns)
            out_params = TypeInfo(var_pair)

        else:
            _out_variable_name = kparams.pop('_out_variable_name',
                                '%s%s' % (func_name, spyne.const.RESULT_SUFFIX))

            out_params[_out_variable_name] = _returns

    ns = spyne.const.xml.DEFAULT_NS
    if _out_message_name.startswith("{"):
        _out_message_name_parts = _out_message_name[1:].partition("}")
        ns = _out_message_name_parts[0]  # skip index 1, it is the closing '}'
        _out_message_name = _out_message_name_parts[2]

    if body_style_str.endswith('bare') and _returns is not None:
        message = _returns.customize(sub_name=_out_message_name, sub_ns=ns)
        if message.__type_name__ is ModelBase.Empty:
            message.__type_name__ = _out_message_name

    else:
        message = ComplexModel.produce(type_name=_out_message_name,
                                               namespace=ns, members=out_params)

        message.Attributes._wrapper = True
        message.__namespace__ = ns  # FIXME: is this necessary?

    if _out_wsdl_part_name:
        message = message.customize(wsdl_part_name=_out_wsdl_part_name)

    return message


def _substitute_self_reference(params, kparams, self_ref_replacement, _no_self):
    from spyne.model import SelfReference

    for i, v in enumerate(params):
        if isclass(v) and issubclass(v, SelfReference):
            if _no_self:
                raise LogicError("SelfReference can't be used in @rpc")
            params[i] = recust_selfref(v, self_ref_replacement)
        else:
            params[i] = v

    for k, v in kparams.items():
        if isclass(v) and issubclass(v, SelfReference):
            if _no_self:
                raise LogicError("SelfReference can't be used in @rpc")
            kparams[k] = recust_selfref(v, self_ref_replacement)
        else:
            kparams[k] = v


def _get_event_managers(kparams):
    _evmgr = kparams.pop("_evmgr", None)
    _evmgrs = kparams.pop("_evmgrs", None)

    if _evmgr is not None and _evmgrs is not None:
        raise LogicError("Pass one of _evmgr or _evmgrs but not both")

    if _evmgr is not None:
        _evmgrs = [_evmgr]

    _event_manager = kparams.pop("_event_manager", None)
    _event_managers = kparams.pop("_event_managers", None)

    if _event_manager is not None and _event_managers is not None:
        raise LogicError("Pass one of _event_manager or "
                         "_event_managers but not both")

    if _event_manager is not None:
        _event_managers = [_event_manager]

    if _evmgrs is not None and _event_managers is not None:
        raise LogicError("You must pass at most one of _evmgr* "
                         "arguments or _event_manager* arguments")
    elif _evmgrs is not None:
        _event_managers = _evmgrs

    return _event_managers if _event_managers is not None else []


def rpc(*params, **kparams):
    """Method decorator to tag a method as a remote procedure call in a
    :class:`spyne.service.Service` subclass.

    You should use the :class:`spyne.server.null.NullServer` transport if you
    want to call the methods directly. You can also use the 'function' attribute
    of the returned object to call the function itself.

    ``_operation_name`` vs ``_in_message_name``:
    Soap clients(SoapUI, Savon, suds) will use the operation name as the
    function name. The name of the input message(_in_message_name) is irrelevant
    when interfacing in this manner; this is because the clients mostly wrap
    around it. However, the soap xml request only uses the input message when
    posting with the soap server; the other protocols only use the input message
    as well. ``_operation_name`` cannot be used with ``_in_message_name``.

    :param _returns: Denotes The return type of the function. It can be a
        type or a sequence of types for functions that have multiple return
        values.
    :param _in_header: A type or an iterable of types that that this method
        accepts as incoming header.
    :param _out_header: A type or an iterable of types that that this method
        sends as outgoing header.
    :param _operation_name: The function's soap operation name. The operation
        and SoapAction names will be equal to the value of ``_operation_name``.
        Default is the function name.
    :param _in_message_name: The public name of the function's input message.
        Default is: ``_operation_name + REQUEST_SUFFIX``.
    :param _out_message_name: The public name of the function's output message.
        Default is: ``_operation_name + RESPONSE_SUFFIX``.
    :param _in_arg_names: The public names of the function arguments. It's
        a dict that maps argument names in the code to public ones.
    :param _in_variable_names: **DEPRECATED** Same as _in_arg_names, kept for
        backwards compatibility.
    :param _out_variable_name: The public name of the function response object.
        It's a string. Ignored when ``_body_style != 'wrapped'`` or ``_returns``
        is a sequence.
    :param _out_variable_names: The public name of the function response object.
        It's a sequence of strings. Ignored when ``_body_style != 'wrapped'`` or
        or ``_returns`` is not a sequence. Must be the same length as
        ``_returns``.
    :param _body_style: One of ``('bare', 'wrapped')``. Default: ``'wrapped'``.
        In wrapped mode, wraps response objects in an additional class.
    :param _soap_body_style: One of ('rpc', 'document'). Default ``'document'``.
        ``_soap_body_style='document'`` is an alias for
        ``_body_style='wrapped'``. ``_soap_body_style='rpc'`` is an alias for
        ``_body_style='bare'``.
    :param _port_type: Soap port type string.
    :param _no_ctx: Don't pass implicit ctx object to the user method.
    :param _no_self: This method does not get an implicit 'self' argument
        (before any other argument, including ctx).
    :param _udd: Short for User Defined Data, you can use this to mark the
        method with arbitrary metadata.
    :param _udp: **DEPRECATED** synonym of ``_udd``.
    :param _aux: The auxiliary backend to run this method. ``None`` if primary.
    :param _throws: A sequence of exceptions that this function can throw. This
        has no real functionality besides publishing this information in
        interface documents.
    :param _args: the name of the arguments to expose.
    :param _event_managers: An iterable of :class:`spyne.EventManager`
        instances. This is useful for adding additional event handlers to
        individual functions.
    :param _event_manager: An instance of :class:`spyne.EventManager` class.
    :param _logged: May be the string '...' to denote that the rpc arguments
        will not be logged.
    :param _evmgrs: Same as ``_event_managers``.
    :param _evmgr: Same as ``_event_manager``.
    :param _service_class: A :class:`Service` subclass. It's generally not a
        good idea to override it for ``@rpc`` methods. It could be necessary to
        override it for ``@mrpc`` methods to add events and other goodies.
    :param _service: Same as ``_service``.
    :param _wsdl_part_name: Overrides the part name attribute within wsdl
        input/output messages eg "parameters"
    """

    params = list(params)

    def explain(f):
        def explain_method(**kwargs):
            # params and kparams are passed by the user to the @rpc family
            # of decorators.

            # kwargs is passed by spyne while sanitizing methods. it mainly
            # contains information about the method context like the service
            # class that contains the method at hand.

            function_name = kwargs['_default_function_name']
            _service_class = kwargs.pop("_service_class", None)
            _self_ref_replacement = None

            # this block is passed straight to the descriptor
            _is_callback = kparams.pop('_is_callback', False)
            _is_async = kparams.pop('_is_async', False)
            _mtom = kparams.pop('_mtom', False)
            _in_header = kparams.pop('_in_header', None)
            _out_header = kparams.pop('_out_header', None)
            _port_type = kparams.pop('_port_type', None)
            _no_ctx = kparams.pop('_no_ctx', False)
            _aux = kparams.pop('_aux', None)
            _pattern = kparams.pop("_pattern", None)
            _patterns = kparams.pop("_patterns", [])
            _args = kparams.pop("_args", None)
            _translations = kparams.pop("_translations", None)
            _when = kparams.pop("_when", None)
            _static_when = kparams.pop("_static_when", None)
            _href = kparams.pop("_href", None)
            _logged = kparams.pop("_logged", True)
            _internal_key_suffix = kparams.pop('_internal_key_suffix', '')
            if '_service' in kparams and '_service_class' in kparams:
                raise LogicError("Please pass only one of '_service' and "
                                                             "'_service_class'")
            if '_service' in kparams:
                _service_class = kparams.pop("_service")
            if '_service_class' in kparams:
                _service_class = kparams.pop("_service_class")

            _no_self = kparams.pop('_no_self', True)
            _event_managers = _get_event_managers(kparams)

            # mrpc-specific
            _self_ref_replacement = kwargs.pop('_self_ref_replacement', None)
            _default_on_null = kparams.pop('_default_on_null', False)
            _substitute_self_reference(params, kparams, _self_ref_replacement,
                                                                       _no_self)

            _faults = None
            if ('_faults' in kparams) and ('_throws' in kparams):
                raise ValueError("only one of '_throws ' or '_faults' arguments"
                                 "must be given -- they're synonyms.")

            elif '_faults' in kparams:
                _faults = kparams.pop('_faults')

            elif '_throws' in kparams:
                _faults = kparams.pop('_throws')

            _is_in_message_name_overridden = not ('_in_message_name' in kparams)
            _in_message_name = kparams.pop('_in_message_name', function_name)

            if _no_self is False and _is_in_message_name_overridden:
                _in_message_name = '%s.%s' % \
                       (_self_ref_replacement.get_type_name(), _in_message_name)

            _operation_name = kparams.pop('_operation_name', function_name)

            if _operation_name != function_name and \
                                              _in_message_name != function_name:
                raise ValueError(
                    "only one of '_operation_name' and '_in_message_name' "
                    "arguments should be given")

            if _in_message_name == function_name:
                _in_message_name = add_request_suffix(_operation_name)

            if '_in_arg_names' in kparams and '_in_variable_names' in kparams:
                raise LogicError("Use either '_in_arg_names' or "
                                              "'_in_variable_names', not both.")
            elif '_in_arg_names' in kparams:
                _in_arg_names = kparams.pop('_in_arg_names')

            elif '_in_variable_names' in kparams:
                _in_arg_names = kparams.pop('_in_variable_names')

            else:
                _in_arg_names = {}

            if '_udd' in kparams and '_udp' in kparams:
                raise LogicError("Use either '_udd' or '_udp', not both.")
            elif '_udd' in kparams:
                _udd = kparams.pop('_udd')

            elif '_udp' in kparams:
                _udd = kparams.pop('_udp')

            else:
                _udd = {}

            _wsdl_part_name = kparams.get('_wsdl_part_name', None)

            body_style = BODY_STYLE_WRAPPED
            body_style_str = _validate_body_style(kparams)
            if body_style_str.endswith('bare'):
                if body_style_str == 'out_bare':
                    body_style = BODY_STYLE_OUT_BARE
                else:
                    body_style = BODY_STYLE_BARE

            in_message = _produce_input_message(f, params,
                    _in_message_name, _in_arg_names, _no_ctx, _no_self,
                                   _args, body_style_str, _self_ref_replacement,
                                   _wsdl_part_name)

            out_message = _produce_output_message(function_name,
                       body_style_str, _self_ref_replacement, _no_self, kparams)

            if _logged != True:
                in_message.Attributes.logged = _logged
                out_message.Attributes.logged = _logged

            doc = getattr(f, '__doc__')

            if _pattern is not None and _patterns != []:
                raise ValueError("only one of '_pattern' and '_patterns' "
                                 "arguments should be given")

            if _pattern is not None:
                _patterns = [_pattern]

            if body_style_str.endswith('bare'):
                from spyne.model import ComplexModelBase

                ti = in_message
                to = out_message
                if issubclass(ti, ComplexModelBase) and len(ti._type_info) == 0:
                    if not issubclass(to, ComplexModelBase) or \
                                                         len(to._type_info) > 0:
                        body_style = BODY_STYLE_EMPTY_OUT_BARE
                    else:
                        body_style = BODY_STYLE_EMPTY

            assert _in_header is None or isinstance(_in_header, tuple)
            retval = MethodDescriptor(f,
                in_message, out_message, doc,
                is_callback=_is_callback, is_async=_is_async, mtom=_mtom,
                in_header=_in_header, out_header=_out_header, faults=_faults,
                parent_class=_self_ref_replacement,
                port_type=_port_type, no_ctx=_no_ctx, udd=_udd,
                class_key=function_name, aux=_aux, patterns=_patterns,
                body_style=body_style, args=_args,
                operation_name=_operation_name, no_self=_no_self,
                translations=_translations,
                when=_when, static_when=_static_when,
                service_class=_service_class, href=_href,
                internal_key_suffix=_internal_key_suffix,
                default_on_null=_default_on_null,
                event_managers=_event_managers,
                logged=_logged,
            )

            if _patterns is not None and _no_self:
                for p in _patterns:
                    p.hello(retval)

            if len(kparams) > 0:
                raise ValueError("Unknown kwarg(s) %r passed.", kparams)
            return retval

        explain_method.__doc__ = f.__doc__
        explain_method._is_rpc = True

        return explain_method

    return explain


def srpc(*params, **kparams):
    """Method decorator to tag a method as a remote procedure call. See
    :func:`spyne.decorator.rpc` for detailed information.

    The initial "s" stands for "static". In Spyne terms, that means no implicit
    first argument is passed to the user callable, which really means the
    method is "stateless" rather than static. It's meant to be used for
    existing functions that can't be changed.
    """

    kparams["_no_ctx"] = True
    return rpc(*params, **kparams)


def mrpc(*params, **kparams):
    kparams["_no_self"] = False
    return rpc(*params, **kparams)



#!/usr/bin/env python
# encoding: utf8
#
# Copyright © BJ Cardon <bj dot car dot don at gmail dot com>,
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#    1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#    3. Neither the name of the owner nor the names of its contributors may be
#       used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
# OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

from django.core.exceptions import ValidationError
from django.db.utils import IntegrityError
from django.views.decorators.csrf import csrf_exempt

from spyne.error import ResourceNotFoundError, ResourceAlreadyExistsError
from spyne.server.django import DjangoApplication
from spyne.model.primitive import Unicode, Integer
from spyne.model.complex import Iterable
from spyne.service import Service
from spyne.protocol.soap import Soap11
from spyne.application import Application
from spyne.decorator import rpc
from spyne.util.django import DjangoComplexModel, DjangoService

from rpctest.core.models import FieldContainer


class Container(DjangoComplexModel):
    class Attributes(DjangoComplexModel.Attributes):
        django_model = FieldContainer
        django_exclude = ['excluded_field']


class HelloWorldService(Service):
    @rpc(Unicode, Integer, _returns=Iterable(Unicode))
    def say_hello(ctx, name, times):
        for i in range(times):
            yield 'Hello, %s' % name


class ContainerService(Service):
    @rpc(Integer, _returns=Container)
    def get_container(ctx, pk):
        try:
            return FieldContainer.objects.get(pk=pk)
        except FieldContainer.DoesNotExist:
            raise ResourceNotFoundError('Container')

    @rpc(Container, _returns=Container)
    def create_container(ctx, container):
        try:
            return FieldContainer.objects.create(**container.as_dict())
        except IntegrityError:
            raise ResourceAlreadyExistsError('Container')

class ExceptionHandlingService(DjangoService):

    """Service for testing exception handling."""

    @rpc(_returns=Container)
    def raise_does_not_exist(ctx):
        return FieldContainer.objects.get(pk=-1)

    @rpc(_returns=Container)
    def raise_validation_error(ctx):
        raise ValidationError(None, 'Invalid.')


app = Application([HelloWorldService, ContainerService,
                   ExceptionHandlingService],
    'spyne.examples.django',
    in_protocol=Soap11(validator='lxml'),
    out_protocol=Soap11(),
)

hello_world_service = csrf_exempt(DjangoApplication(app))



#!/usr/bin/env python
#
# spyne - Copyright (C) Spyne contributors.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
#


import unittest

from spyne import Application, Service, rpc
from spyne.model import Array, ComplexModel, AnyXml, UnsignedLong, \
    UnsignedInteger16, Integer, DateTime, Unicode
from spyne.protocol.http import HttpRpc
from spyne.protocol.soap import Soap11


class TestInterface(unittest.TestCase):
    def test_imports(self):
        import logging
        logging.basicConfig(level=logging.DEBUG)

        class KeyValuePair(ComplexModel):
            __namespace__ = "1"
            key = Unicode
            value = Unicode

        class Something(ComplexModel):
            __namespace__ = "2"
            d = DateTime
            i = Integer

        class SomethingElse(ComplexModel):
            __namespace__ = "3"
            a = AnyXml
            b = UnsignedLong
            s = Something

        class BetterSomething(Something):
            __namespace__ = "4"
            k = UnsignedInteger16

        class Service1(Service):
            @rpc(SomethingElse, _returns=Array(KeyValuePair))
            def some_call(ctx, sth):
                pass

        class Service2(Service):
            @rpc(BetterSomething, _returns=Array(KeyValuePair))
            def some_other_call(ctx, sth):
                pass

        application = Application([Service1, Service2],
            in_protocol=HttpRpc(),
            out_protocol=Soap11(),
            name='Service', tns='target_namespace'
        )

        imports = application.interface.imports
        tns = application.interface.get_tns()
        smm = application.interface.service_method_map
        print(imports)

        assert imports[tns] == set(['1', '3', '4'])
        assert imports['3'] == set(['2'])
        assert imports['4'] == set(['2'])

        assert smm['{%s}some_call' % tns]
        assert smm['{%s}some_call' % tns][0].service_class == Service1
        assert smm['{%s}some_call' % tns][0].function == Service1.some_call

        assert smm['{%s}some_other_call' % tns]
        assert smm['{%s}some_other_call' % tns][0].service_class == Service2
        assert smm['{%s}some_other_call' % tns][0].function == Service2.some_other_call

    def test_custom_primitive_in_array(self):
        RequestStatus = Unicode(values=['new', 'processed'], zonta='bonta')

        class DataRequest(ComplexModel):
            status = Array(RequestStatus)

        class HelloWorldService(Service):
            @rpc(DataRequest)
            def some_call(ctx, dgrntcl):
                pass

        Application([HelloWorldService], 'spyne.examples.hello.soap',
                in_protocol=Soap11(validator='lxml'),
                out_protocol=Soap11())

        # test passes if instantiating Application doesn't fail


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# encoding: utf8
#
# Copyright © Burak Arslan <burak at arskom dot com dot tr>,
#             Arskom Ltd. http://www.arskom.com.tr
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#    1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#    3. Neither the name of the owner nor the names of its contributors may be
#       used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
# OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import logging
logging.basicConfig(level=logging.DEBUG)
logging.getLogger('spyne.protocol.xml').setLevel(logging.DEBUG)
logging.getLogger('sqlalchemy.engine.base.Engine').setLevel(logging.DEBUG)

from spyne import Application, rpc, Mandatory as M, Unicode, Array, Iterable, \
    TTableModel, Service, ResourceNotFoundError, UnsignedInteger32

from spyne.protocol.http import HttpRpc
from spyne.protocol.yaml import YamlDocument

from spyne.util import memoize

from spyne.server.wsgi import WsgiApplication

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker


db = create_engine('sqlite:///:memory:')
Session = sessionmaker(bind=db)
TableModel = TTableModel()
TableModel.Attributes.sqla_metadata.bind = db


class Permission(TableModel):
    __tablename__ = 'permission'
    __namespace__ = 'spyne.examples.sql_crud'
    __table_args__ = {"sqlite_autoincrement": True}

    id = UnsignedInteger32(primary_key=True)
    application = Unicode(256)
    operation = Unicode(256)


class User(TableModel):
    __tablename__ = 'user'
    __namespace__ = 'spyne.examples.sql_crud'
    __table_args__ = {"sqlite_autoincrement": True}

    id = UnsignedInteger32(primary_key=True)
    name = Unicode(256)
    first_name = Unicode(256)
    last_name = Unicode(256)
    permissions = Array(Permission, store_as='table')


@memoize
def TCrudService(T, T_name):
    class CrudService(Service):
        @rpc(M(UnsignedInteger32, _returns=T))
        def get(ctx, obj_id):
            return ctx.udc.session.query(T).filter_by(id=obj_id).one()

        @rpc(T, _returns=UnsignedInteger32)
        def put(ctx, obj):
            if obj.id is None:
                ctx.udc.session.add(obj)
                ctx.udc.session.flush()

            else:
                if ctx.udc.session.query(T).get(obj.id) is None:
                    raise ResourceNotFoundError('%s.id=%d' % (T_name, obj.id))

                else:
                    ctx.udc.session.merge(obj)

            return obj.id

        @rpc(M(UnsignedInteger32))
        def del_(ctx, obj_id):
            count = ctx.udc.session.query(T).filter_by(id=obj_id).count()
            if count == 0:
                raise ResourceNotFoundError(obj_id)

            ctx.udc.session.query(T).filter_by(id=obj_id).delete()

        @rpc(_returns=Iterable(T))
        def get_all(ctx):
            return ctx.udc.session.query(T)

    return CrudService


class UserDefinedContext(object):
    def __init__(self):
        self.session = Session()


def _on_method_call(ctx):
    ctx.udc = UserDefinedContext()


def _on_method_return_object(ctx):
    ctx.udc.session.commit()


def _on_method_context_closed(ctx):
    if ctx.udc is not None:
        ctx.udc.session.close()


user_service = TCrudService(User, 'user')
application = Application(
    [user_service],
    tns='spyne.examples.sql_crud',
    in_protocol=HttpRpc(validator='soft'),
    out_protocol=YamlDocument()
)


application.event_manager.add_listener('method_call', _on_method_call)
application.event_manager.add_listener('method_return_object',
                                                      _on_method_return_object)
application.event_manager.add_listener("method_context_closed",
                                                      _on_method_context_closed)


if __name__ == '__main__':
    from wsgiref.simple_server import make_server

    wsgi_app = WsgiApplication(application)
    server = make_server('127.0.0.1', 8000, wsgi_app)

    TableModel.Attributes.sqla_metadata.create_all()
    logging.info("listening to http://127.0.0.1:8000")
    logging.info("wsdl is at: http://localhost:8000/?wsdl")

    server.serve_forever()



# encoding: utf8
#
# spyne - Copyright (C) Spyne contributors.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
#

from __future__ import absolute_import

import logging
logger = logging.getLogger(__name__)

import msgpack

from mmap import mmap
from collections import OrderedDict

from spyne import MethodContext, TransportContext, Address
from spyne.auxproc import process_contexts
from spyne.error import ValidationError, InternalError
from spyne.server import ServerBase
from spyne.util.six import binary_type

try:
    from twisted.internet.defer import Deferred
except ImportError as e:
    def Deferred(*_, **__): raise e


MSGPACK_SHELL_OVERHEAD = 10


def _process_v1_msg(prot, msg):
    header = None
    body = msg[1]
    if not isinstance(body, (binary_type, mmap, memoryview)):
        raise ValidationError(body, "Body must be a bytestream.")

    if len(msg) > 2:
        header = msg[2]
        if not isinstance(header, dict):
            raise ValidationError(header, "Header must be a dict.")
        for k, v in header.items():
            header[k] = msgpack.unpackb(v)

    ctx = MessagePackMethodContext(prot, MessagePackMethodContext.SERVER)
    ctx.in_string = [body]
    ctx.transport.in_header = header

    return ctx


class MessagePackTransportContext(TransportContext):
    def __init__(self, parent, transport):
        super(MessagePackTransportContext, self).__init__(parent, transport)

        self.in_header = None
        self.protocol = None
        self.inreq_queue = OrderedDict()
        self.request_len = None

    def get_peer(self):
        if self.protocol is not None:
            peer = self.protocol.transport.getPeer()
            return Address.from_twisted_address(peer)


class MessagePackOobMethodContext(object):
    __slots__ = 'd'

    def __init__(self):
        if Deferred is not None:
            self.d = Deferred()
        else:
            self.d = None

    def close(self):
        if self.d is not None and not self.d.called:
            self.d.cancel()


class MessagePackMethodContext(MethodContext):
    TransportContext = MessagePackTransportContext

    def __init__(self, transport, way):
        self.oob_ctx = None

        super(MessagePackMethodContext, self).__init__(transport, way)

    def close(self):
        super(MessagePackMethodContext, self).close()
        if self.transport is not None:
            self.transport.protocol = None
            self.transport = None

        if self.oob_ctx is not None:
            self.oob_ctx.close()


class MessagePackTransportBase(ServerBase):
    # These are all placeholders that need to be overridden in subclasses
    OUT_RESPONSE_NO_ERROR = None
    OUT_RESPONSE_CLIENT_ERROR = None
    OUT_RESPONSE_SERVER_ERROR = None

    IN_REQUEST = None

    def __init__(self, app):
        super(MessagePackTransportBase, self).__init__(app)

        self._version_map = {
            self.IN_REQUEST: _process_v1_msg
        }

    def produce_contexts(self, msg):
        """Produce contexts based on incoming message.

        :param msg: Parsed request in this format: `[IN_REQUEST, body, header]`
        """

        if not isinstance(msg, (list, tuple)):
            logger.debug("Incoming request: %r", msg)
            raise ValidationError(msg, "Request must be a list")

        if not len(msg) >= 2:
            logger.debug("Incoming request: %r", msg)
            raise ValidationError(len(msg), "Request must have at least two "
                                                          "elements. It has %r")

        if not isinstance(msg[0], int):
            logger.debug("Incoming request: %r", msg)
            raise ValidationError(msg[0], "Request version must be an integer. "
                                                                    "It was %r")

        processor = self._version_map.get(msg[0], None)
        if processor is None:
            logger.debug("Invalid incoming request: %r", msg)
            raise ValidationError(msg[0], "Unknown request type %r")

        msglen = len(msg[1])
        # shellen = len(msgpack.packb(msg))
        # logger.debug("Shell size: %d, message size: %d, diff: %d",
        #                                     shellen, msglen, shellen - msglen)
        # some approx. msgpack overhead based on observations of what's above.
        msglen += MSGPACK_SHELL_OVERHEAD

        initial_ctx = processor(self, msg)
        contexts = self.generate_contexts(initial_ctx)

        p_ctx, others = contexts[0], contexts[1:]
        p_ctx.transport.request_len = msglen

        return p_ctx, others

    def process_contexts(self, contexts):
        p_ctx, others = contexts[0], contexts[1:]

        if p_ctx.in_error:
            return self.handle_error(p_ctx, others, p_ctx.in_error)

        self.get_in_object(p_ctx)
        if p_ctx.in_error:
            logger.error(p_ctx.in_error)
            return self.handle_error(p_ctx, others, p_ctx.in_error)

        self.get_out_object(p_ctx)
        if p_ctx.out_error:
            return self.handle_error(p_ctx, others, p_ctx.out_error)

        try:
            self.get_out_string(p_ctx)

        except Exception as e:
            logger.exception(e)
            contexts.out_error = InternalError("Serialization Error.")
            return self.handle_error(contexts, others, contexts.out_error)

    def handle_error(self, p_ctx, others, error):
        self.get_out_string(p_ctx)

        try:
            process_contexts(self, others, p_ctx, error=error)
        except Exception as e:
            # Report but ignore any exceptions from auxiliary methods.
            logger.exception(e)

    def handle_transport_error(self, error):
        return msgpack.dumps(str(error))

    def pack(self, ctx):
        ctx.out_string = msgpack.packb([self.OUT_RESPONSE_NO_ERROR,
                                                     b''.join(ctx.out_string)]),


class MessagePackServerBase(MessagePackTransportBase):
    """Contains the transport protocol logic but not the transport itself.

    Subclasses should implement logic to move bitstreams in and out of this
    class."""

    OUT_RESPONSE_NO_ERROR = 0
    OUT_RESPONSE_CLIENT_ERROR = 1
    OUT_RESPONSE_SERVER_ERROR = 2

    IN_REQUEST = 1



#!/usr/bin/env python

import logging
logger = logging.getLogger(__name__)

import os
logging.basicConfig(level=logging.DEBUG)
logging.getLogger('spyne.protocol.xml').setLevel(logging.DEBUG)

from spyne.application import Application
from spyne.decorator import rpc
from spyne.service import ServiceBase
from spyne.model.complex import ComplexModel
from spyne.model.binary import ByteArray
from spyne.model.primitive import Unicode
from spyne.server.wsgi import WsgiApplication
from spyne.protocol.soap import Soap12


tns = 'http://gib.gov.tr/vedop3/eFatura'


class documentResponse(ComplexModel):
    msg = Unicode
    hash = ByteArray


class GIBSoapService(ServiceBase):
    @rpc(Unicode(sub_name="fileName"), ByteArray(sub_name='binaryData'),
                          ByteArray(sub_name="hash"), _returns=documentResponse)
    def documentRequest(ctx, file_name, file_data, data_hash):
        incoming_invoice_dir = os.getcwd()

        logger.info("file_name  %r" % file_name)
        logger.info("file_hash: %r" % data_hash)

        path = os.path.join(incoming_invoice_dir, file_name)

        f = open(path, 'wb')
        for data in file_data:
            f.write(data)
        logger.info("File written: %r" % file_name)
        f.close()

        resp = documentResponse()
        resp.msg = "Document was written successfully"
        resp.hash = data_hash

        return resp


application = Application([GIBSoapService], tns=tns,
                          in_protocol=Soap12(),
                          out_protocol=Soap12())


gib_application = WsgiApplication(application)


from wsgiref.simple_server import make_server


server = make_server('0.0.0.0', 8000, gib_application)
server.serve_forever()



# encoding: utf8
#
# spyne - Copyright (C) Spyne contributors.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
#

from __future__ import print_function

import logging
logger = logging.getLogger(__name__)

from inspect import isgenerator

from lxml.html.builder import E

from spyne import ModelBase, ByteArray, ComplexModelBase, Array, AnyUri, \
    ImageUri
from spyne.util import coroutine, Break
from spyne.util.cdict import cdict
from spyne.protocol.html.table import HtmlTableBase


class HtmlRowTable(HtmlTableBase):
    """Protocol that returns the response object as a html table.

    The simple flavour is like the HtmlMicroFormatprotocol, but returns data
    as a html table using the <table> tag.

    Returns one record per table in a table with two columns.

    :param app: A spyne.application.Application instance.
    :param header: Boolean value to determine whether to show field
        names in the beginning of the table or not. Defaults to True. Set to
        False to skip headers.
    :param table_name_attr: The name of the attribute that will contain the
        response name of the complex object in the table tag. Set to None to
        disable.
    :param table_name: When not none, overrides what goes in `table_name_attr`.
    :param table_class: When not none, specifies what goes in `class` attribute
        in the `<table>` tag. Table name gets appended when
        `table_name_attr == 'class'`
    :param field_name_attr: The name of the attribute that will contain the
        field names of the complex object children for every table cell. Set
        to None to disable.
    :param row_class: value that goes inside the <tr class="">
    :param cell_class: value that goes inside the <td class="">
    :param header_cell_class: value that goes inside the <th class="">
    """

    def __init__(self, *args, **kwargs):
        super(HtmlRowTable, self).__init__(*args, **kwargs)

        self.serialization_handlers = cdict({
            ModelBase: self.model_base_to_parent,
            AnyUri: self.any_uri_to_parent,
            ImageUri: self.imageuri_to_parent,
            ByteArray: self.not_supported,
            ComplexModelBase: self.complex_model_to_parent,
            Array: self.array_to_parent,
        })

    def model_base_to_parent(self, ctx, cls, inst, parent, name, from_arr=False,
                                                                      **kwargs):
        if from_arr:
            td_attrib = {}
            if False and self.field_name_attr:
                td_attrib[self.field_name_attr] = name

            parent.write(E.tr(E.td(self.to_unicode(cls, inst), **td_attrib)))
        else:
            parent.write(self.to_unicode(cls, inst))

    @coroutine
    def complex_model_to_parent(self, ctx, cls, inst, parent, name,
                                                      from_arr=False, **kwargs):
        attrib = {}
        if self.table_name_attr is not None:
            attrib[self.table_name_attr] = cls.get_type_name()
        if self.table_width is not None:
            attrib['width'] = self.table_width

        with parent.element('table', attrib):
            with parent.element('tbody'):
                for k, v in self.sort_fields(cls):
                    sub_attrs = self.get_cls_attrs(v)
                    if sub_attrs.exc:
                        logger.debug("\tExclude table cell %r type %r for %r",
                                                                      k, v, cls)
                        continue
                    try:
                        sub_value = getattr(inst, k, None)
                    except:  # e.g. SQLAlchemy could throw NoSuchColumnError
                        sub_value = None

                    sub_name = v.Attributes.sub_name
                    if sub_name is None:
                        sub_name = k

                    tr_attrs = {}
                    if self.row_class is not None:
                        self.add_html_attr('class', tr_attrs, self.row_class)

                    with parent.element('tr', tr_attrs):
                        th_attrs = {}

                        if self.header_cell_class is not None:
                            self.add_html_attr('class', th_attrs,
                                                         self.header_cell_class)

                        self.add_field_attrs(th_attrs, sub_name, v)

                        if sub_attrs.hidden:
                            self.add_style(th_attrs, 'display:None')

                        if self.header:
                            parent.write(E.th(
                                self.trc(v, ctx.locale, sub_name),
                                **th_attrs
                            ))

                        td_attrs = {}
                        if self.cell_class is not None:
                            self.add_html_attr('class', td_attrs,
                                                                self.cell_class)

                        self.add_field_attrs(td_attrs, sub_name, v)

                        if sub_attrs.hidden:
                            self.add_style(td_attrs, 'display:None')

                        with parent.element('td', td_attrs):
                            ret = self.to_parent(ctx, v, sub_value, parent,
                                                             sub_name, **kwargs)
                            if isgenerator(ret):
                                try:
                                    while True:
                                        sv2 = (yield)
                                        ret.send(sv2)
                                except Break as b:
                                    try:
                                        ret.throw(b)
                                    except StopIteration:
                                        pass

    @coroutine
    def array_to_parent(self, ctx, cls, inst, parent, name, **kwargs):
        with parent.element('div'):
            if issubclass(cls, ComplexModelBase):
                ret = super(HtmlRowTable, self).array_to_parent(
                                         ctx, cls, inst, parent, name, **kwargs)
                if isgenerator(ret):
                    try:
                        while True:
                            sv2 = (yield)
                            ret.send(sv2)
                    except Break as b:
                        try:
                            ret.throw(b)
                        except StopIteration:
                            pass
            else:
                table_attrib = {}
                if self.table_name_attr:
                    table_attrib = {self.table_name_attr: name}
                if self.table_width is not None:
                    table_attrib['width'] = self.table_width

                with parent.element('table', table_attrib):
                    tr_attrib = {}
                    if self.row_class is not None:
                        tr_attrib['class'] = self.row_class
                    with parent.element('tr', tr_attrib):
                        if self.header:
                            parent.write(E.th(self.trc(cls, ctx.locale,
                                                          cls.get_type_name())))
                        td_attrs = {}

                        if self.cell_class is not None:
                            self.add_html_attr('class', td_attrs,
                                                                self.cell_class)

                        self.add_field_attrs(td_attrs, name, cls)

                        cls_attrs = self.get_cls_attrs(cls)

                        if cls_attrs.hidden:
                            self.add_style(td_attrs, 'display:None')

                        with parent.element('td', td_attrs):
                            with parent.element('table'):
                                ret = super(HtmlRowTable, self) \
                                    .array_to_parent(ctx, cls, inst, parent,
                                                                 name, **kwargs)
                                if isgenerator(ret):
                                    try:
                                        while True:
                                            sv2 = (yield)
                                            ret.send(sv2)
                                    except Break as b:
                                        try:
                                            ret.throw(b)
                                        except StopIteration:
                                            pass



# encoding: utf8
#
# Copyright © Burak Arslan <burak at arskom dot com dot tr>,
#             Arskom Ltd. http://www.arskom.com.tr
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#    1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#    3. Neither the name of the owner nor the names of its contributors may be
#       used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
# OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import logging
logger = logging.getLogger(__name__)

from spyne.application import Application
from spyne.error import Fault
from spyne.error import InternalError
from spyne.error import ResourceNotFoundError
from spyne.util.email import email_exception

from sqlalchemy.orm.exc import NoResultFound

from template.context import UserDefinedContext

EXCEPTION_ADDRESS = "everybody@example.com"


def _on_method_call(ctx):
    ctx.udc = UserDefinedContext()


def _on_method_context_closed(ctx):
    if ctx.udc is not None:
        ctx.udc.session.commit()
        ctx.udc.session.close()


class MyApplication(Application):
    def __init__(self, services, tns, name=None,
                                         in_protocol=None, out_protocol=None):
        super(MyApplication, self).__init__(services, tns, name, in_protocol,
                                                                 out_protocol)

        self.event_manager.add_listener('method_call', _on_method_call)
        self.event_manager.add_listener("method_context_closed",
                                                    _on_method_context_closed)

    def call_wrapper(self, ctx):
        try:
            return ctx.service_class.call_wrapper(ctx)

        except NoResultFound:
            raise ResourceNotFoundError(ctx.in_object)

        except Fault, e:
            logger.error(e)
            raise

        except Exception, e:
            logger.exception(e)
            # This should not happen! Let the team know via email
            email_exception(EXCEPTION_ADDRESS)
            raise InternalError(e)



#!/usr/bin/env python3

from __future__ import print_function, unicode_literals

import sys

from lxml import etree
from spyne.interface.xml_schema.parser import hier_repr

from spyne.util import six
from spyne import ComplexModel, Unicode
from spyne.util.xml import get_object_as_xml_polymorphic, \
    get_xml_as_object_polymorphic


# uncomment to see what's going on under the hood
# import logging
# logging.basicConfig(level=logging.DEBUG,
#     format='%(levelname)-7s %(module)12s:%(lineno)-4d | %(message)s')


class B(ComplexModel):
    __namespace__ = 'some_ns'
    _type_info = {
        '_b': Unicode,
    }

    def __init__(self):
        super(B, self).__init__()
        self._b = "b"


class C(B):
    __namespace__ = 'some_ns'
    _type_info = {
        '_c': Unicode,
    }

    def __init__(self):
        super(C, self).__init__()
        self._c = "c"


class A(ComplexModel):
    __namespace__ = 'some_ns'
    _type_info = {
        '_a': Unicode,
        '_b': B,
    }

    def __init__(self, b=None):
        super(A, self).__init__()
        self._a = 'a'
        self._b = b


a = A(b=C())
elt = get_object_as_xml_polymorphic(a, A)
xml_string = etree.tostring(elt, pretty_print=True)
if six.PY2:
    print(xml_string, end="")
else:
    sys.stdout.buffer.write(xml_string)

element_tree = etree.fromstring(xml_string)
new_a = get_xml_as_object_polymorphic(elt, A)

assert new_a._a == a._a, (a._a, new_a._a)
assert new_a._b._b == a._b._b, (a._b._b, new_a._b._b)
assert new_a._b._c == a._b._c, (a._b._c, new_a._b._c)

print(hier_repr(new_a))




#
# spyne - Copyright (C) Spyne contributors.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
#

"""The ``spyne.model.complex`` module contains
:class:`spyne.model.complex.ComplexBase` class and its helper objects and
subclasses. These are mainly container classes for other simple or
complex objects -- they don't carry any data by themselves.
"""

from __future__ import print_function

import logging
logger = logging.getLogger(__name__)

import decimal
import traceback

from copy import copy
from weakref import WeakKeyDictionary
from collections import deque, OrderedDict
from inspect import isclass
from itertools import chain

from spyne import const
from spyne.const.xml import PREFMAP

from spyne.model import Point, Unicode, PushBase, ModelBase
from spyne.model._base import PSSM_VALUES, apply_pssm
from spyne.model.primitive import NATIVE_MAP
from spyne.model.primitive._base import AnyXml

from spyne.util import six, memoize, memoize_id, sanitize_args, \
    memoize_ignore_none
from spyne.util.color import YEL
from spyne.util.meta import Prepareable
from spyne.util.odict import odict
from spyne.util.six import add_metaclass, with_metaclass, string_types

# FIXME: for backwards compatibility, to be removed in Spyne 3
# noinspection PyUnresolvedReferences
from spyne.model import json, jsonb, xml, msgpack, table


def _get_flat_type_info(cls, retval):
    assert isinstance(retval, TypeInfo)
    parent = getattr(cls, '__extends__', None)
    if not (parent is None):
        _get_flat_type_info(parent, retval)
    retval.update(cls._type_info)
    retval.alt.update(cls._type_info_alt)  # FIXME: move to cls._type_info.alt
    retval.attrs.update({k: v for (k, v) in cls._type_info.items()
                                                if issubclass(v, XmlAttribute)})
    return retval


class TypeInfo(odict):
    def __init__(self, *args, **kwargs):
        super(TypeInfo, self).__init__(*args, **kwargs)

        self.attributes = {}
        self.alt = OrderedDict()
        self.attrs = OrderedDict()

    def __setitem__(self, key, val):
        assert isinstance(key, string_types)
        super(TypeInfo, self).__setitem__(key, val)


class _SimpleTypeInfoElement(object):
    __slots__ = ['path', 'parent', 'type', 'is_array', 'can_be_empty']

    def __init__(self, path, parent, type_, is_array, can_be_empty):
        self.path = path
        self.parent = parent
        self.type = type_
        self.is_array = is_array
        self.can_be_empty = can_be_empty

    def __repr__(self):
        return "SimpleTypeInfoElement(path=%r, parent=%r, type=%r, is_array=%r)" \
                            % (self.path, self.parent, self.type, self.is_array)


class XmlModifier(ModelBase):
    def __new__(cls, type, ns=None):
        retval = cls.customize()
        retval.type = type
        retval.Attributes = type.Attributes
        retval._ns = ns
        if type.__type_name__ is ModelBase.Empty:
            retval.__type_name__ = ModelBase.Empty
        return retval

    @staticmethod
    def resolve_namespace(cls, default_ns, tags=None):
        cls.type.resolve_namespace(cls.type, default_ns, tags)

        cls.__namespace__ = cls._ns

        if cls.__namespace__ is None:
            cls.__namespace__ = cls.type.get_namespace()

        if cls.__namespace__ in PREFMAP:
            cls.__namespace__ = default_ns

    @classmethod
    def _fill_empty_type_name(cls, parent_ns, parent_tn, k):
        cls.__namespace__ = parent_ns
        tn = "%s_%s%s" % (parent_tn, k, const.TYPE_SUFFIX)

        child_v = cls.type
        child_v.__type_name__ = tn

        cls._type_info = TypeInfo({tn: child_v})
        cls.__type_name__ = '%s%s%s' % (const.ARRAY_PREFIX, tn,
                                                         const.ARRAY_SUFFIX)

        extends = child_v.__extends__
        while extends is not None and extends.get_type_name() is cls.Empty:
            extends._fill_empty_type_name(parent_ns, parent_tn,
                                                    k + const.PARENT_SUFFIX)
            extends = extends.__extends__


class XmlData(XmlModifier):
    """Items which are marshalled as data of the parent element."""

    @classmethod
    def marshall(cls, prot, name, value, parent_elt):
        if value is not None:
            if issubclass(cls.type, AnyXml):
                parent_elt.append(value)
            else:
                if len(parent_elt) == 0:
                    parent_elt.text = prot.to_bytes(cls.type, value)
                else:
                    parent_elt[-1].tail = prot.to_bytes(cls.type, value)

    @classmethod
    def get_type_name(cls):
        return cls.type.get_type_name()

    @classmethod
    def get_type_name_ns(cls, interface):
        return cls.type.get_type_name_ns(interface)

    @classmethod
    def get_namespace(cls):
        return cls.type.get_namespace()

    @classmethod
    def get_element_name(cls):
        return cls.type.get_element_name()

    @classmethod
    def get_element_name_ns(cls, interface):
        return cls.type.get_element_name_ns(interface)


class XmlAttribute(XmlModifier):
    """Items which are marshalled as attributes of the parent element."""

    def __new__(cls, type_, use=None, ns=None):
        retval = super(XmlAttribute, cls).__new__(cls, type_, ns)
        retval._use = use
        if retval.type.Attributes.min_occurs > 0 and retval._use is None:
            retval._use = 'required'
        return retval


class XmlAttributeRef(XmlAttribute):
    """Reference to an Xml attribute."""

    def __init__(self, ref, use=None):
        self._ref = ref
        self._use = use

    def describe(self, name, element, app):
        element.set('ref', self._ref)
        if self._use:
            element.set('use', self._use)


class SelfReference(object):
    """Use this as a placeholder type in classes that contain themselves. See
    :func:`spyne.test.model.test_complex.TestComplexModel.test_self_reference`.
    """
    customize_args = []
    customize_kwargs = {}
    __orig__ = None

    def __init__(self):
        raise NotImplementedError()

    @classmethod
    def customize(cls, *args, **kwargs):
        args = list(chain(args, cls.customize_args))
        kwargs = dict(chain(kwargs.items(), cls.customize_kwargs.items()))
        if cls.__orig__ is None:
            cls.__orig__ = cls

        return type("SelfReference", (cls,), {
            'customize_args': args,
            'customize_kwargs': kwargs,
        })


def _get_spyne_type(cls_name, k, v):
    try:
        v = NATIVE_MAP.get(v, v)
    except TypeError:
        return

    try:
        subc = issubclass(v, ModelBase) or issubclass(v, SelfReference)
    except:
        subc = False

    if subc:
        if issubclass(v, Array) and len(v._type_info) != 1:
            raise Exception("Invalid Array definition in %s.%s."% (cls_name, k))
        elif issubclass(v, Point) and v.Attributes.dim is None:
            raise Exception("Please specify the number of dimensions")
        return v


def _join_args(x, y):
    if x is None:
        return y
    if y is None:
        return x

    xa, xk = sanitize_args(x)
    ya, yk = sanitize_args(y)

    xk = dict(xk)
    xk.update(yk)

    return xa + ya, xk


def _gen_attrs(cls_bases, cls_dict):
    attrs = cls_dict.get('Attributes', None)
    if attrs is None:
        for b in cls_bases:
            if hasattr(b, 'Attributes'):
                class Attributes(b.Attributes):
                    pass
                attrs = cls_dict['Attributes'] = Attributes
                break
        else:
            raise Exception("No ModelBase subclass in bases? Huh?")

    return attrs


def _get_type_info(cls, cls_name, cls_bases, cls_dict, attrs):
    base_type_info = TypeInfo()
    mixin = TypeInfo()
    extends = cls_dict.get('__extends__', None)

    # user did not specify explicit base class so let's try to derive it from
    # the actual class hierarchy
    if extends is None:
        # we don't want origs end up as base classes
        orig = cls_dict.get("__orig__", None)
        if orig is None:
            orig = getattr(cls, '__orig__', None)

        if orig is not None:
            bases = orig.__bases__
            logger.debug("Got bases for %s from orig: %r", cls_name, bases)
        else:
            bases = cls_bases
            logger.debug("Got bases for %s from meta: %r", cls_name, bases)

        for b in bases:
            base_types = getattr(b, "_type_info", None)

            # we don't care about non-ComplexModel bases
            if base_types is None:
                continue

            # mixins are simple
            if getattr(b, '__mixin__', False) == True:
                logger.debug("Adding fields from mixin %r to '%s'", b, cls_name)
                mixin.update(b.get_flat_type_info(b))

                if '__mixin__' not in cls_dict:
                    cls_dict['__mixin__'] = False

                continue

            if not (extends in (None, b)):
                raise Exception("Spyne objects do not support multiple "
                    "inheritance. Use mixins if you need to reuse "
                    "fields from multiple classes.")

            if len(base_types) > 0 and issubclass(b, ModelBase):
                extends = cls_dict["__extends__"] = b
                assert extends.__orig__ is None, "You can't inherit from a " \
                    "customized class. You should first get your class " \
                    "hierarchy right, then start customizing classes."

                b.get_subclasses.memo.clear()
                logger.debug("Registering %r as base of '%s'", b, cls_name)

    if not ('_type_info' in cls_dict):
        cls_dict['_type_info'] = _type_info = TypeInfo()
        _type_info.update(base_type_info)

        class_fields = []
        for k, v in cls_dict.items():
            if k.startswith('_'):
                continue

            if isinstance(v, tuple) and len(v) == 1 and \
                                 _get_spyne_type(cls_name, k, v[0]) is not None:
                logger.warning(YEL("There seems to be a stray comma in the"
                                   "definition of '%s.%s'.", cls_name, k))

            v = _get_spyne_type(cls_name, k, v)

            if v is None:
                continue

            class_fields.append((k, v))

        _type_info.update(class_fields)

    else:
        _type_info = cls_dict['_type_info']

        if not isinstance(_type_info, TypeInfo):
            _type_info = cls_dict['_type_info'] = TypeInfo(_type_info)

    for k, v in reversed(mixin.items()):
        _type_info.insert(0, (k, v))

    return _type_info


class _MethodsDict(dict):
    def __init__(self, *args, **kwargs):
        super(_MethodsDict, self).__init__(*args, **kwargs)

        self._processed = False


def _gen_methods(cls, cls_dict):
    methods = _MethodsDict()
    for k, v in cls_dict.items():
        if not k.startswith('_') and hasattr(v, '_is_rpc'):
            logger.debug("Registering %s as member method for %r", k, cls)
            assert cls is not None

            # generate method descriptor from information in the decorator
            descriptor = v(_default_function_name=k, _self_ref_replacement=cls)

            # strip the decorator and put the original function in the class
            setattr(cls, k, descriptor.function)

            # modify the descriptor with user-supplied class
            if cls.Attributes.method_config_do is not None:
                descriptor = cls.Attributes.method_config_do(descriptor)

            methods[k] = descriptor

    return methods


def _get_ordered_attributes(cls_name, cls_dict, attrs):
    if not isinstance(cls_dict, odict):
        # FIXME: Maybe add a warning here?
        return cls_dict

    SUPPORTED_ORDERS = ('random', 'declared')
    if (attrs.declare_order is not None and
            not attrs.declare_order in SUPPORTED_ORDERS):

        msg = "The declare_order attribute value %r is invalid in %s"
        raise Exception(msg % (attrs.declare_order, cls_name))

    declare_order = attrs.declare_order or const.DEFAULT_DECLARE_ORDER
    if declare_order is None or declare_order == 'random':
        # support old behaviour
        cls_dict = dict(cls_dict)

    return cls_dict


def _sanitize_sqlalchemy_parameters(cls_dict, attrs):
    table_name = cls_dict.get('__tablename__', None)
    if attrs.table_name is None:
        attrs.table_name = table_name

    _cls_table = cls_dict.get('__table__', None)
    if attrs.sqla_table is None:
        attrs.sqla_table = _cls_table

    metadata = cls_dict.get('__metadata__', None)
    if attrs.sqla_metadata is None:
        attrs.sqla_metadata = metadata

    margs = cls_dict.get('__mapper_args__', None)
    attrs.sqla_mapper_args = _join_args(attrs.sqla_mapper_args, margs)

    targs = cls_dict.get('__table_args__', None)
    attrs.sqla_table_args = _join_args(attrs.sqla_table_args, targs)


def _sanitize_type_info(cls_name, _type_info, _type_info_alt):
    """Make sure _type_info contents are sane"""

    for k, v in _type_info.items():
        if not isinstance(k, six.string_types):
            raise ValueError("Invalid class key", k)

        if not isclass(v):
            raise ValueError(v)

        if issubclass(v, SelfReference):
            continue

        elif not issubclass(v, ModelBase):
            v = _get_spyne_type(cls_name, k, v)
            if v is None:
                raise ValueError( (cls_name, k, v) )
            _type_info[k] = v

        elif issubclass(v, Array) and len(v._type_info) != 1:
            raise Exception("Invalid Array definition in %s.%s." %
                                                        (cls_name, k))
        sub_ns = v.Attributes.sub_ns
        sub_name = v.Attributes.sub_name

        if sub_ns is None and sub_name is None:
            pass

        elif sub_ns is not None and sub_name is not None:
            key = "{%s}%s" % (sub_ns, sub_name)
            if key in _type_info:
                raise Exception("%r is already defined: %r" %
                                                         (key, _type_info[key]))
            _type_info_alt[key] = v, k

        elif sub_ns is None:
            key = sub_name
            if sub_ns in _type_info:
                raise Exception("%r is already defined: %r" %
                                                         (key, _type_info[key]))
            _type_info_alt[key] = v, k

        elif sub_name is None:
            key = "{%s}%s" % (sub_ns, k)
            if key in _type_info:
                raise Exception("%r is already defined: %r" %
                                                         (key, _type_info[key]))
            _type_info_alt[key] = v, k


D_EXC = dict(exc=True)


def _process_child_attrs(cls, retval, kwargs):
    child_attrs = copy(kwargs.get('child_attrs', None))
    child_attrs_all = kwargs.get('child_attrs_all', None)
    child_attrs_noexc = copy(kwargs.get('child_attrs_noexc', None))

    # add exc=False to child_attrs_noexc
    if child_attrs_noexc is not None:
        # if there is _noexc, make sure that child_attrs_all is also used to
        # exclude exclude everything else first
        if child_attrs_all is None:
            child_attrs_all = D_EXC

        else:
            if 'exc' in child_attrs_all and child_attrs_all['exc'] != D_EXC:
                logger.warning("Overriding child_attrs_all['exc'] to True "
                                                                  "for %r", cls)

            child_attrs_all.update(D_EXC)

        # update child_attrs_noexc with exc=False
        for k, v in child_attrs_noexc.items():
            if 'exc' in v:
                logger.warning("Overriding 'exc' for %s.%s from "
                         "child_attrs_noexc with False", cls.get_type_name(), k)

            v['exc'] = False

        # update child_attrs with data from child_attrs_noexc
        if child_attrs is None:
            child_attrs = child_attrs_noexc

        else:
            # update with child_attrs_noexc with exc=False
            if child_attrs is None:
                child_attrs = dict()

            for k, v in child_attrs_noexc.items():
                if k in child_attrs:
                    logger.warning("Overriding child_attrs for %s.%s from "
                                    "child_attrs_noexc", cls.get_type_name(), k)

                child_attrs[k] = v

    if child_attrs_all is not None:
        ti = retval._type_info
        logger.debug("processing child_attrs_all for %r", cls)
        for k, v in ti.items():
            logger.debug("  child_attrs_all set %r=%r", k, child_attrs_all)
            ti[k] = ti[k].customize(**child_attrs_all)

        if retval.__extends__ is not None:
            retval.__extends__ = retval.__extends__.customize(
                                                child_attrs_all=child_attrs_all)

        retval.Attributes._delayed_child_attrs_all = child_attrs_all

    if child_attrs is not None:
        ti = retval._type_info
        logger.debug("processing child_attrs for %r", cls)
        for k, v in list(child_attrs.items()):
            if k in ti:
                logger.debug("  child_attr set %r=%r", k, v)
                ti[k] = ti[k].customize(**v)
                del child_attrs[k]

        base_fti = {}
        if retval.__extends__ is not None:
            retval.__extends__ = retval.__extends__.customize(
                                                        child_attrs=child_attrs)
            base_fti = retval.__extends__.get_flat_type_info(retval.__extends__)

        for k, v in child_attrs.items():
            if k not in base_fti:
                logger.debug("  child_attr delayed %r=%r", k, v)
                retval.Attributes._delayed_child_attrs[k] = v


def recust_selfref(selfref, cls):
    if len(selfref.customize_args) > 0 or len(selfref.customize_kwargs) > 0:
        logger.debug("Replace self reference with %r with *%r and **%r",
                          cls, selfref.customize_args, selfref.customize_kwargs)
        return cls.customize(*selfref.customize_args,
                                                     **selfref.customize_kwargs)
    logger.debug("Replace self reference with %r", cls)
    return cls


def _set_member_default(inst, key, cls, attr):
    def_val = attr.default
    def_fac = attr.default_factory

    if def_fac is None and def_val is None:
        return False

    if def_fac is not None:
        if six.PY2 and hasattr(def_fac, 'im_func'):
            # unbound-method error workaround. huh.
            def_fac = def_fac.im_func

        dval = def_fac()

        # should not check for read-only for default values
        setattr(inst, key, dval)

        return True

    if def_val is not None:
        # should not check for read-only for default values
        setattr(inst, key, def_val)

        return True

    assert False, "Invalid application state"


def _is_sqla_array(cls, attr):
    # inner object is complex
    ret1 = issubclass(cls, Array) and \
                              hasattr(cls.get_inner_type(), '_sa_class_manager')

    # inner object is primitive
    ret2 = issubclass(cls, Array) and attr.store_as is not None

    # object is a bare array
    ret3 = attr.max_occurs > 1 and hasattr(cls, '_sa_class_manager')

    return ret1 or ret2 or ret3


def _init_member(inst, key, cls, attr):
    cls_getattr_ret = getattr(inst.__class__, key, None)

    if isinstance(cls_getattr_ret, property) and cls_getattr_ret.fset is None:
        return  # we skip read-only properties

    if _set_member_default(inst, key, cls, attr):
       return

    # sqlalchemy objects do their own init.
    if _is_sqla_array(cls, attr):
        # except the attributes that sqlalchemy doesn't know about
        if attr.exc_db:
            setattr(inst, key, None)

        elif attr.store_as is None:
            setattr(inst, key, None)

        return

    # sqlalchemy objects do their own init.
    if hasattr(inst.__class__, '_sa_class_manager'):
        # except the attributes that sqlalchemy doesn't know about
        if attr.exc_db:
            setattr(inst, key, None)

        elif issubclass(cls, ComplexModelBase) and attr.store_as is None:
            setattr(inst, key, None)

        return

    setattr(inst, key, None)


class ComplexModelMeta(with_metaclass(Prepareable, type(ModelBase))):
    """This metaclass sets ``_type_info``, ``__type_name__`` and ``__extends__``
    which are going to be used for (de)serialization and schema generation.
    """

    def __new__(cls, cls_name, cls_bases, cls_dict):
        """This function initializes the class and registers attributes."""

        attrs = _gen_attrs(cls_bases, cls_dict)
        assert issubclass(attrs, ComplexModelBase.Attributes), \
                   ("%r must be a ComplexModelBase.Attributes subclass" % attrs)

        cls_dict = _get_ordered_attributes(cls_name, cls_dict, attrs)

        type_name = cls_dict.get("__type_name__", None)
        if type_name is None:
            cls_dict["__type_name__"] = cls_name

        _type_info = _get_type_info(cls, cls_name, cls_bases, cls_dict, attrs)

        # used for sub_name and sub_ns
        _type_info_alt = cls_dict['_type_info_alt'] = TypeInfo()
        for b in cls_bases:
            if hasattr(b, '_type_info_alt'):
                _type_info_alt.update(b._type_info_alt)

        _sanitize_type_info(cls_name, _type_info, _type_info_alt)
        _sanitize_sqlalchemy_parameters(cls_dict, attrs)

        return super(ComplexModelMeta, cls).__new__(cls,
                                                  cls_name, cls_bases, cls_dict)

    def __init__(self, cls_name, cls_bases, cls_dict):
        type_info = self._type_info

        extends = self.__extends__
        if extends is not None and self.__orig__ is None:
            eattr = extends.Attributes
            if eattr._subclasses is None:
                eattr._subclasses = []
            eattr._subclasses.append(self)
            if self.Attributes._subclasses is eattr._subclasses:
                self.Attributes._subclasses = None

        # sanitize fields
        for k, v in type_info.items():
            # replace bare SelfRerefence
            if issubclass(v, SelfReference):
                self._replace_field(k, recust_selfref(v, self))

            # cache XmlData for easier access
            elif issubclass(v, XmlData):
                if self.Attributes._xml_tag_body_as is None:
                    self.Attributes._xml_tag_body_as = [(k, v)]
                else:
                    self.Attributes._xml_tag_body_as.append((k, v))

            # replace SelfRerefence in arrays
            elif issubclass(v, Array):
                v2, = v._type_info.values()
                while issubclass(v2, Array):
                    v = v2
                    v2, = v2._type_info.values()

                if issubclass(v2, SelfReference):
                    v._set_serializer(recust_selfref(v2, self))

        # apply field order
        # FIXME: Implement this better
        new_type_info = []
        for k, v in self._type_info.items():
            if v.Attributes.order == None:
                new_type_info.append(k)

        for k, v in self._type_info.items():
            if v.Attributes.order is not None:
                new_type_info.insert(v.Attributes.order, k)

        assert len(self._type_info) == len(new_type_info)
        self._type_info.keys()[:] = new_type_info

        # install checkers for validation on assignment
        for k, v in self._type_info.items():
            if not v.Attributes.validate_on_assignment:
                continue

            def _get_prop(self):
                return self.__dict__[k]

            def _set_prop(self, val):
                if not (val is None or isinstance(val, v.Value)):
                    raise ValueError("Invalid value %r, "
                                 "should be an instance of %r" % (val, v.Value))

                self.__dict__[k] = val

            setattr(self, k, property(_get_prop, _set_prop))

        # process member rpc methods
        methods = _gen_methods(self, cls_dict)
        if len(methods) > 0:
            self.Attributes.methods = methods

        # finalize sql table mapping
        tn = self.Attributes.table_name
        meta = self.Attributes.sqla_metadata
        t = self.Attributes.sqla_table

        # For spyne objects reflecting an existing db table
        if tn is None:
            if t is not None:
                self.Attributes.sqla_metadata = t.metadata
                from spyne.store.relational import gen_spyne_info

                gen_spyne_info(self)

        # For spyne objects being converted to a sqlalchemy table
        elif meta is not None and (tn is not None or t is not None) and \
                                                       len(self._type_info) > 0:
            from spyne.store.relational import gen_sqla_info

            gen_sqla_info(self, cls_bases)

        super(ComplexModelMeta, self).__init__(cls_name, cls_bases, cls_dict)

    #
    # We record the order fields are defined into ordered dict, so we can
    # declare them in the same order in the WSDL.
    #
    # For Python 3 __prepare__ works out of the box, see PEP 3115.
    # But we use `Preparable` metaclass for both Python 2 and Python 3 to
    # support six.add_metaclass decorator
    #
    @classmethod
    def __prepare__(mcs, name, bases, **kwds):
        return odict()


_is_array = lambda v: issubclass(v, Array) or (v.Attributes.max_occurs > 1)


class ComplexModelBase(ModelBase):
    """If you want to make a better class type, this is what you should inherit
    from.
    """

    __mixin__ = False

    class Attributes(ModelBase.Attributes):
        """ComplexModel-specific attributes"""

        store_as = None
        """Method for serializing to persistent storage. One of %r. It makes
        sense to specify this only when this object is a child of another
        ComplexModel subclass.""" % (PSSM_VALUES,)

        sqla_metadata = None
        """None or :class:`sqlalchemy.MetaData` instance."""

        sqla_table_args = None
        """A dict that will be passed to :class:`sqlalchemy.schema.Table`
        constructor as ``**kwargs``.
        """

        sqla_mapper_args = None
        """A dict that will be passed to :func:`sqlalchemy.orm.mapper`
        constructor as. ``**kwargs``.
        """

        sqla_table = None
        """The sqlalchemy table object"""

        sqla_mapper = None
        """The sqlalchemy mapper object"""

        validate_freq = True
        """When ``False``, soft validation ignores missing mandatory attributes.
        """

        child_attrs = None
        """Customize child attributes in one go. It's a dict of dicts. This is
        ignored unless used via explicit customization."""

        child_attrs_all = None
        """Customize all child attributes. It's a dict. This is ignored unless
        used via explicit customization. `child_attrs` always take precedence.
        """

        declare_order = None
        """The order fields of the :class:``ComplexModel`` are to be declared
        in the SOAP WSDL. If this is left as None or explicitly set to
        ``'random'`` declares then the fields appear in whatever order the
        Python's hash map implementation seems fit in the WSDL. This randomised
        order can change every time the program is run. This is what Spyne <2.11
        did if you didn't set _type_info as an explicit sequence (e.g. using a
        list, odict, etc.). It means that clients who are manually complied or
        generated from the WSDL will likely need to be recompiled every time it
        changes. The string ``name`` means the field names are alphabetically
        sorted in the WSDL declaration.  The string ``declared`` means in the
        order the field type was declared in Python 2, and the order the
        field was declared in Python 3.

        In order to get declared field order in Python 2, the
        :class:`spyne.util.meta.Preparable` class inspects the frame stack in
        order to locate the class definition, re-parses it to get declaration
        order from the AST and uses that information to order elements.

        It's a horrible hack that we tested to work with CPython 2.6 through 3.3
        and PyPy. It breaks in Nuitka as Nuitka does away with code objects.
        Other platforms were not tested.

        It's not recommended to use set this to ``'declared'`` in Python 2
        unless you're sure you fully understand the consequences.
        """

        parent_variant = None
        """FIXME: document me yo."""

        methods = None
        """A dict of member RPC methods (typically marked with @mrpc)."""

        method_config_do = None
        """When not None, it's a callable that accepts a ``@mrpc`` method
        descriptor and returns a modified version."""

        not_wrapped = None
        """When True, serializes to non-wrapped object, overriding the protocol
        flag."""

        wrapped = None
        """When True, serializes to a wrapped object, overriding the protocol
        flag. When a str/bytes/unicode value, uses that value as key wrapper
        object name."""

        _variants = None
        _xml_tag_body_as = None
        _delayed_child_attrs = None
        _delayed_child_attrs_all = None
        _subclasses = None

    def __init__(self, *args, **kwargs):
        cls = self.__class__
        cls_attr = cls.Attributes
        fti = cls.get_flat_type_info(cls)

        if cls.__orig__ is not None:
            logger.warning("%r(0x%X) seems to be a customized class. It is not "
                    "supposed to be instantiated. You have been warned.",
                                                                   cls, id(cls))
            logger.debug(traceback.format_stack())

        if cls_attr._xml_tag_body_as is not None:
            for arg, (xtba_key, xtba_type) in \
                                           zip(args, cls_attr._xml_tag_body_as):

                if xtba_key is not None and len(args) == 1:
                    attr = xtba_type.Attributes
                    _init_member(self, xtba_key, xtba_type, attr)
                    self._safe_set(xtba_key, arg, xtba_type,
                                                           xtba_type.Attributes)
                elif len(args) > 0:
                    raise TypeError(
                                "Positional argument is only for ComplexModels "
                                "with XmlData field. You must use keyword "
                                "arguments in any other case.")

        for k, v in fti.items():
            attr = v.Attributes
            if not k in self.__dict__:
                _init_member(self, k, v, attr)

            if k in kwargs:
                self._safe_set(k, kwargs[k], v, attr)

    def __len__(self):
        return len(self._type_info)

    def __getitem__(self, i):
        if isinstance(i, slice):
            retval = []
            for key in self._type_info.keys()[i]:
                retval.append(getattr(self, key, None))

        else:
            retval = getattr(self, self._type_info.keys()[i], None)

        return retval

    def __repr__(self):
        return "%s(%s)" % (self.get_type_name(), ', '.join(
               ['%s=%r' % (k, self.__dict__.get(k))
                    for k in self.__class__.get_flat_type_info(self.__class__)
                    if self.__dict__.get(k, None) is not None]))

    def _safe_set(self, key, value, t, attrs):
        if attrs.read_only:
            return False

        try:
            setattr(self, key, value)
        except AttributeError as e:
            logger.exception(e)
            raise AttributeError("can't set %r attribute %s to %r" %
                                                   (self.__class__, key, value))

        return True

    @classmethod
    def get_identifiers(cls):
        for k, v in cls.get_flat_type_info(cls).items():
            if getattr(v.Attributes, 'primary_key', None):
                yield k, v

    @classmethod
    def get_primary_keys(cls):
        return cls.get_identifiers()

    def as_dict(self):
        """Represent object as dict.

        Null values are omitted from dict representation to support optional
        not nullable attributes.
        """

        return dict((
            (k, getattr(self, k)) for k in self.get_flat_type_info(self.__class__)
            if getattr(self, k) is not None
        ))

    @classmethod
    def get_serialization_instance(cls, value):
        """Returns the native object corresponding to the serialized form passed
        in the ``value`` argument.

        :param value: This argument can be:
            * A list or tuple of native types aligned with cls._type_info.
            * A dict of native types.
            * The native type itself.

            If the value type is not a ``list``, ``tuple`` or ``dict``, the
            value is returned untouched.
        """

        # if the instance is a list, convert it to a cls instance.
        # this is only useful when deserializing method arguments for a client
        # request which is the only time when the member order is not arbitrary
        # (as the members are declared and passed around as sequences of
        # arguments, unlike dictionaries in a regular class definition).
        if isinstance(value, list) or isinstance(value, tuple):
            keys = cls.get_flat_type_info(cls).keys()

            if not len(value) <= len(keys):
                logger.error("\n\tcls: %r" "\n\tvalue: %r" "\n\tkeys: %r",
                                                               cls, value, keys)
                raise ValueError("Impossible sequence to instance conversion")

            cls_orig = cls
            if cls.__orig__ is not None:
                cls_orig = cls.__orig__

            try:
                inst = cls_orig()

            except Exception as e:
                logger.error("Error instantiating %r: %r", cls_orig, e)
                raise

            for i in range(len(value)):
                setattr(inst, keys[i], value[i])

        elif isinstance(value, dict):
            cls_orig = cls
            if cls.__orig__ is not None:
                cls_orig = cls.__orig__
            inst = cls_orig()

            for k in cls.get_flat_type_info(cls):
                setattr(inst, k, value.get(k, None))

        else:
            inst = value

        return inst

    @classmethod
    def get_deserialization_instance(cls, ctx):
        """Get an empty native type so that the deserialization logic can set
        its attributes.
        """
        if cls.__orig__ is None:
            return cls()
        return cls.__orig__()

    @classmethod
    @memoize_id
    def get_subclasses(cls):
        retval = []
        subca = cls.Attributes._subclasses
        if subca is not None:
            retval.extend(subca)
            for subc in subca:
                retval.extend(subc.get_subclasses())
        return retval

    @staticmethod
    @memoize_ignore_none
    def get_flat_type_info(cls):
        """Returns a _type_info dict that includes members from all base
        classes.

        It's called a "flat" dict because it flattens all members from the
        inheritance hierarchy into one dict.
        """
        return _get_flat_type_info(cls, TypeInfo())

    @classmethod
    def get_orig(cls):
        return cls.__orig__ or cls

    @staticmethod
    def get_simple_type_info(cls, hier_delim="."):
        """Returns a _type_info dict that includes members from all base classes
        and whose types are only primitives. It will prefix field names in
        non-top-level complex objects with field name of its parent.

        For example, given hier_delim='.'; the following hierarchy: ::

            {'some_object': [{'some_string': ['abc']}]}

        would be transformed to: ::

            {'some_object.some_string': ['abc']}

        :param hier_delim: String that will be used as delimiter between field
            names. Default is ``'.'``.
        """
        return ComplexModelBase.get_simple_type_info_with_prot(
                                                     cls, hier_delim=hier_delim)

    @staticmethod
    @memoize
    def get_simple_type_info_with_prot(cls, prot=None, hier_delim="."):
        """See :func:ComplexModelBase.get_simple_type_info"""
        fti = cls.get_flat_type_info(cls)

        retval = TypeInfo()
        tags = set()

        queue = deque()
        if prot is None:
            for k, v in fti.items():
                sub_name = k

                queue.append((
                    (k,),
                    v,
                    (sub_name,),
                    (_is_array(v),),
                    cls,
                ))

        else:
            for k, v in fti.items():
                cls_attrs = prot.get_cls_attrs(v)
                sub_name = cls_attrs.sub_name
                if sub_name is None:
                    sub_name = k

                queue.append((
                    (k,),
                    v,
                    (sub_name,),
                    (_is_array(v),),
                    cls,
                ))

        tags.add(cls)

        while len(queue) > 0:
            keys, v, prefix, is_array, parent = queue.popleft()
            k = keys[-1]
            if issubclass(v, Array) and v.Attributes.max_occurs == 1:
                v, = v._type_info.values()

            key = hier_delim.join(prefix)
            if issubclass(v, ComplexModelBase):
                retval[key] = _SimpleTypeInfoElement(
                    path=keys,
                    parent=parent,
                    type_=v,
                    is_array=tuple(is_array),
                    can_be_empty=True,
                )

                if not (v in tags):
                    tags.add(v)
                    if prot is None:
                        for k2, v2 in v.get_flat_type_info(v).items():
                            sub_name = k2
                            queue.append((
                                keys + (k2,),
                                v2,
                                prefix + (sub_name,),
                                is_array + (_is_array(v),),
                                v
                            ))

                    else:
                        for k2, v2 in v.get_flat_type_info(v).items():
                            cls_attrs = prot.get_cls_attrs(v2)
                            sub_name = cls_attrs.sub_name
                            if sub_name is None:
                                sub_name = k2

                            queue.append((
                                keys + (k2,),
                                v2,
                                prefix + (sub_name,),
                                is_array + (_is_array(v),),
                                v,
                            ))

            else:
                value = retval.get(key, None)

                if value is not None:
                    raise ValueError("%r.%s conflicts with %r" %
                                                       (cls, k, value.path))

                retval[key] = _SimpleTypeInfoElement(
                    path=keys,
                    parent=parent,
                    type_=v,
                    is_array=tuple(is_array),
                    can_be_empty=False,
                )

        return retval

    @staticmethod
    def resolve_namespace(cls, default_ns, tags=None):
        if tags is None:
            tags = set()
        elif cls in tags:
            return False

        if not ModelBase.resolve_namespace(cls, default_ns, tags):
            return False

        for k, v in cls._type_info.items():
            if v is None:
                continue

            if v.__type_name__ is ModelBase.Empty:
                v._fill_empty_type_name(cls.get_namespace(),
                                                         cls.get_type_name(), k)

            v.resolve_namespace(v, default_ns, tags)

        if cls._force_own_namespace is not None:
            for c in cls._force_own_namespace:
                c.__namespace__ = cls.get_namespace()
                ComplexModel.resolve_namespace(c, cls.get_namespace(), tags)

        assert not (cls.__namespace__ is ModelBase.Empty)
        assert not (cls.__type_name__ is ModelBase.Empty)

        return True

    @staticmethod
    def produce(namespace, type_name, members):
        """Lets you create a class programmatically."""

        return ComplexModelMeta(type_name, (ComplexModel,), odict({
            '__namespace__': namespace,
            '__type_name__': type_name,
            '_type_info': TypeInfo(members),
        }))

    @classmethod
    def customize(cls, **kwargs):
        """Duplicates cls and overwrites the values in ``cls.Attributes`` with
        ``**kwargs`` and returns the new class.

        Because each class is registered as a variant of the original (__orig__)
        class, using this function to generate classes dynamically on-the-fly
        could cause memory leaks. You have been warned.
        """

        store_as = apply_pssm(kwargs.get('store_as', None))
        if store_as is not None:
            kwargs['store_as'] = store_as

        cls_name, cls_bases, cls_dict = cls._s_customize(**kwargs)
        cls_dict['__module__'] = cls.__module__
        if '__extends__' not in cls_dict:
            cls_dict['__extends__'] = cls.__extends__

        retval = type(cls_name, cls_bases, cls_dict)
        retval._type_info = TypeInfo(cls._type_info)
        retval.__type_name__ = cls.__type_name__
        retval.__namespace__ = cls.__namespace__
        retval.Attributes.parent_variant = cls

        dca = retval.Attributes._delayed_child_attrs
        if retval.Attributes._delayed_child_attrs is None:
            retval.Attributes._delayed_child_attrs = {}
        else:
            retval.Attributes._delayed_child_attrs = dict(dca.items())

        tn = kwargs.get("type_name", None)
        if tn is not None:
            retval.__type_name__ = tn

        ns = kwargs.get("namespace", None)
        if ns is not None:
            retval.__namespace__ = ns

        if cls is not ComplexModel:
            cls._process_variants(retval)

        _process_child_attrs(cls, retval, kwargs)

        # we could be smarter, but customize is supposed to be called only
        # during daemon initialization, so it's not really necessary.
        ComplexModelBase.get_subclasses.memo.clear()
        ComplexModelBase.get_flat_type_info.memo.clear()
        ComplexModelBase.get_simple_type_info_with_prot.memo.clear()

        return retval

    @classmethod
    def _process_variants(cls, retval):
        orig = getattr(retval, '__orig__', None)
        if orig is not None:
            if orig.Attributes._variants is None:
                orig.Attributes._variants = WeakKeyDictionary()
            orig.Attributes._variants[retval] = True
            # _variants is only for the root class.
            retval.Attributes._variants = None

    @classmethod
    def _append_field_impl(cls, field_name, field_type):
        assert isinstance(field_name, string_types)

        dcaa = cls.Attributes._delayed_child_attrs_all
        if dcaa is not None:
            field_type = field_type.customize(**dcaa)

        dca = cls.Attributes._delayed_child_attrs
        if dca is not None:
            d_cust = dca.get(field_name, None)
            if d_cust is not None:
                field_type = field_type.customize(**d_cust)

        cls._type_info[field_name] = field_type

        ComplexModelBase.get_flat_type_info.memo.clear()
        ComplexModelBase.get_simple_type_info_with_prot.memo.clear()

    @classmethod
    def _append_to_variants(cls, field_name, field_type):
        if cls.Attributes._variants is not None:
            for c in cls.Attributes._variants:
                c.append_field(field_name, field_type)

    @classmethod
    def append_field(cls, field_name, field_type):
        cls._append_field_impl(field_name, field_type)
        cls._append_to_variants(field_name, field_type)

    @classmethod
    def _insert_to_variants(cls, index, field_name, field_type):
        if cls.Attributes._variants is not None:
            for c in cls.Attributes._variants:
                c.insert_field(index, field_name, field_type)

    @classmethod
    def _insert_field_impl(cls, index, field_name, field_type):
        assert isinstance(index, int)
        assert isinstance(field_name, string_types)

        dcaa = cls.Attributes._delayed_child_attrs_all
        if dcaa is not None:
            field_type = field_type.customize(**dcaa)

        dca = cls.Attributes._delayed_child_attrs
        if dca is not None:
            if field_name in dca:
                d_cust = dca.pop(field_name)
                field_type = field_type.customize(**d_cust)

        cls._type_info.insert(index, (field_name, field_type))

        ComplexModelBase.get_flat_type_info.memo.clear()
        ComplexModelBase.get_simple_type_info_with_prot.memo.clear()

    @classmethod
    def insert_field(cls, index, field_name, field_type):
        cls._insert_field_impl(index, field_name, field_type)
        cls._insert_to_variants(index, field_name, field_type)

    @classmethod
    def _replace_in_variants(cls, field_name, field_type):
        if cls.Attributes._variants is not None:
            for c in cls.Attributes._variants:
                c._replace_field(field_name, field_type)

    @classmethod
    def _replace_field_impl(cls, field_name, field_type):
        assert isinstance(field_name, string_types)

        cls._type_info[field_name] = field_type

        ComplexModelBase.get_flat_type_info.memo.clear()
        ComplexModelBase.get_simple_type_info_with_prot.memo.clear()

    @classmethod
    def _replace_field(cls, field_name, field_type):
        cls._replace_field_impl(field_name, field_type)
        cls._replace_in_variants(field_name, field_type)

    @classmethod
    def store_as(cls, what):
        return cls.customize(store_as=what)

    @classmethod
    def novalidate_freq(cls):
        return cls.customize(validate_freq=False)

    @classmethod
    def init_from(cls, other, **kwargs):
        retval = (cls if cls.__orig__ is None else cls.__orig__)()

        for k, v in cls.get_flat_type_info(cls).items():
            try:
                if k in kwargs:
                    retval._safe_set(k, kwargs[k], v, v.Attributes)

                elif hasattr(other, k):
                    retval._safe_set(k, getattr(other, k), v, v.Attributes)

            except AttributeError as e:
                logger.warning("Error setting %s: %r", k, e)

        return retval

    @classmethod
    def __respawn__(cls, ctx=None, filters=None):
        if ctx is not None and ctx.in_object is not None and \
                                                         len(ctx.in_object) > 0:
            retval = next(iter(ctx.in_object))
            if retval is not None:
                return retval

        if ctx.descriptor.default_on_null:
            return cls.get_deserialization_instance(ctx)


@add_metaclass(ComplexModelMeta)
class ComplexModel(ComplexModelBase):
    """The general complexType factory. The __call__ method of this class will
    return instances, contrary to primivites where the same call will result in
    customized duplicates of the original class definition.
    Those who'd like to customize the class should use the customize method.
    (see :class:``spyne.model.ModelBase``).
    """


@add_metaclass(ComplexModelMeta)
class Array(ComplexModelBase):
    """This class generates a ComplexModel child that has one attribute that has
    the same name as the serialized class. It's contained in a Python list.
    """

    class Attributes(ComplexModelBase.Attributes):
        _wrapper = True

    def __new__(cls, serializer, member_name=None, wrapped=True, **kwargs):
        if not wrapped:
            if serializer.Attributes.max_occurs == 1:
                kwargs['max_occurs'] = 'unbounded'

            return serializer.customize(**kwargs)

        retval = cls.customize(**kwargs)

        _serializer = _get_spyne_type(cls.__name__, '__serializer__', serializer)
        if _serializer is None:
            raise ValueError("serializer=%r is not a valid spyne type" % serializer)

        if issubclass(_serializer, SelfReference):
             # hack to make sure the array passes ComplexModel sanity checks
             # that are there to prevent empty arrays.
            retval._type_info = {'_bogus': _serializer}
        else:
            retval._set_serializer(_serializer, member_name)

        tn = kwargs.get("type_name", None)
        if tn is not None:
            retval.__type_name__ = tn

        return retval

    @classmethod
    def _fill_empty_type_name(cls, parent_ns, parent_tn, k):
        cls.__namespace__ = parent_ns
        tn = "%s_%s%s" % (parent_tn, k, const.TYPE_SUFFIX)

        child_v, = cls._type_info.values()
        child_v.__type_name__ = tn

        cls._type_info = TypeInfo({tn: child_v})
        cls.__type_name__ = '%s%s%s' % (const.ARRAY_PREFIX, tn,
                                                             const.ARRAY_SUFFIX)

        extends = child_v.__extends__
        while extends is not None and extends.get_type_name() is cls.Empty:
            extends._fill_empty_type_name(parent_ns, parent_tn,
                                                    k + const.PARENT_SUFFIX)
            extends = extends.__extends__

    @classmethod
    def customize(cls, **kwargs):
        serializer_attrs = kwargs.get('serializer_attrs', None)
        if serializer_attrs is None:
            return super(Array, cls).customize(**kwargs)

        del kwargs['serializer_attrs']

        logger.debug('Pass serializer attrs %r', serializer_attrs)

        serializer, = cls._type_info.values()
        return cls(serializer.customize(**serializer_attrs)).customize(**kwargs)

    @classmethod
    def _set_serializer(cls, serializer, member_name=None):
        if serializer.get_type_name() is ModelBase.Empty:  # A customized class
            member_name = "OhNoes"
            # mark array type name as "to be resolved later".
            cls.__type_name__ = ModelBase.Empty

        else:
            if member_name is None:
                member_name = serializer.get_type_name()

            cls.__type_name__ = '%s%s%s' % (const.ARRAY_PREFIX,
                                                serializer.get_type_name(),
                                                             const.ARRAY_SUFFIX)

        # hack to default to unbounded arrays when the user didn't specify
        # max_occurs.
        if serializer.Attributes.max_occurs == 1:
            serializer = serializer.customize(max_occurs=decimal.Decimal('inf'))

        assert isinstance(member_name, string_types), member_name
        cls._type_info = TypeInfo({member_name: serializer})

    # the array belongs to its child's namespace, it doesn't have its own
    # namespace.
    @staticmethod
    def resolve_namespace(cls, default_ns, tags=None):
        (serializer,) = cls._type_info.values()

        serializer.resolve_namespace(serializer, default_ns, tags)

        if cls.__namespace__ is None:
            cls.__namespace__ = serializer.get_namespace()

        if cls.__namespace__ in PREFMAP:
            cls.__namespace__ = default_ns

        return ComplexModel.resolve_namespace(cls, default_ns, tags)

    @classmethod
    def get_serialization_instance(cls, value):
        inst = ComplexModel.__new__(Array)

        (member_name,) = cls._type_info.keys()
        setattr(inst, member_name, value)

        return inst

    @classmethod
    def get_deserialization_instance(cls, ctx):
        return []

    @classmethod
    def get_inner_type(cls):
        return next(iter(cls._type_info.values()))


class Iterable(Array):
    """This class generates a ``ComplexModel`` child that has one attribute that
    has the same name as the serialized class. It's contained in a Python
    iterable. The distinction with the ``Array`` is made in the protocol
    implementation, this is just a marker.

    Whenever you return a generator instead of a list, you should use this type
    as this suggests the intermediate machinery to NEVER actually try to iterate
    over the value. An ``Array`` could be iterated over for e.g. logging
    purposes.
    """

    class Attributes(Array.Attributes):
        logged = False

    class Push(PushBase):
        """The push interface to the `Iterable`.

        Anything append()'ed to a `Push` instance is serialized and written to
        outgoing stream immediately.

        When using Twisted, Push callbacks are called from the reactor thread if
        the instantiation is done in a reactor thread. Otherwise, callbacks are
        called by `deferToThread`. Make sure to avoid relying on thread-local
        stuff as `deferToThread` is not guaranteed to restore original thread
        context.
        """
        pass


def TTableModelBase():
    from spyne.store.relational import add_column

    class TableModelBase(ComplexModelBase):
        @classmethod
        def append_field(cls, field_name, field_type):
            cls._append_field_impl(field_name, field_type)
            # There could have been changes to field_type in ComplexModel so we
            # should not use field_type directly from above
            if cls.__table__ is not None:
                add_column(cls, field_name, cls._type_info[field_name])
            cls._append_to_variants(field_name, field_type)

        @classmethod
        def replace_field(cls, field_name, field_type):
            raise NotImplementedError()

        @classmethod
        def insert_field(cls, index, field_name, field_type):
            cls._insert_field_impl(index, field_name, field_type)
            # There could have been changes to field_type in ComplexModel so we
            # should not use field_type directly from above
            if cls.__table__ is not None:
                add_column(cls, field_name, cls._type_info[field_name])
            cls._insert_to_variants(index, field_name, field_type)

    return TableModelBase


# this has docstring repeated in the documentation at reference/model/complex.rst
def TTableModel(metadata=None, base=None, metaclass=None):
    """A TableModel template that generates a new TableModel class for each
    call. If metadata is not supplied, a new one is instantiated.
    """

    from sqlalchemy import MetaData

    if base is None:
        base = TTableModelBase()
    if metaclass is None:
        metaclass = ComplexModelMeta

    @add_metaclass(metaclass)
    class TableModel(base):
        class Attributes(ComplexModelBase.Attributes):
            sqla_metadata = metadata if metadata is not None else MetaData()

    return TableModel


def Mandatory(cls, **_kwargs):
    """Customizes the given type to be a mandatory one. Has special cases for
    :class:`spyne.model.primitive.Unicode` and
    :class:`spyne.model.complex.Array`\\.
    """

    kwargs = dict(min_occurs=1, nillable=False)
    if cls.get_type_name() is not cls.Empty:
        kwargs['type_name'] = '%s%s%s' % (const.MANDATORY_PREFIX,
                                    cls.get_type_name(), const.MANDATORY_SUFFIX)
    kwargs.update(_kwargs)
    if issubclass(cls, Unicode):
        kwargs.update(dict(min_len=1))

    elif issubclass(cls, Array):
        (k,v), = cls._type_info.items()
        if v.Attributes.min_occurs == 0:
            cls._type_info[k] = Mandatory(v)

    return cls.customize(**kwargs)




#
# spyne - Copyright (C) Spyne contributors.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
#

from __future__ import unicode_literals

import decimal
import uuid

from spyne.model.primitive import NATIVE_MAP
from spyne.util import six
from spyne.model._base import SimpleModel
from spyne.model.primitive._base import re_match_with_span


UUID_PATTERN = "%(x)s{8}-%(x)s{4}-%(x)s{4}-%(x)s{4}-%(x)s{12}" % \
                                                            {'x': '[a-fA-F0-9]'}

LTREE_PATTERN = r"\w+(\.\w+)*"

# Actual ltree max size is 65536 but it's advised to keep it under 2048.
LTREE_OPTIMAL_SIZE = 2048
LTREE_MAXIMUM_SIZE = 65536


'[0-9A-Za-z!#$%&\'*+.^_`|~-]+/([0-9A-Za-z!#$%&\'*+.^_`|~-]+);[ \\t]*[0-9A-Za-z!#$%&\'*+.^_`|~-]+=(?:[0-9A-Za-z!#$%&\'*+.^_`|~-]+|"(?:[^"\\\\]|\\.)*");?[ \\t]*([0-9A-Za-z!#$%&\'*+.^_`|~-]+=(?:[0-9A-Za-z!#$%&\'*+.^_`|~-]+|"(?:[^"\\\\]|\\.)*");?[ \\t]*)*'

def _gen_mime_type_pattern(strict, with_params):
    ows = "[ \\t]*"  # Optional WhiteSpace
    token = "[0-9A-Za-z!#$%&'*+.^_`|~-]+"
    quotedString = "\"(?:[^\"\\\\]|\\.)*\""
    if strict:
        main_type = "(" \
                "application|audio|font|example|image|message|model|multipart" \
                "|text|video|x-(?:" + token + ")" \
            ")"

    else:
        main_type = token

    if not with_params:
        return main_type + "/" + "(" + token + ")"

    param = token + "=" + "(?:" + token + "|" + quotedString + ");?" + ows
    params = \
        "(" + ";" \
             + "(" + ows + param + "(" + param + ")*" + ")?" \
        + ")?"

    return main_type + "/" + "(" + token + ")" + params


MIME_TYPE_PATTERN_STRICT = \
    _gen_mime_type_pattern(strict=True, with_params=False)
MIME_TYPE_PATTERN_PERMISSIVE = \
    _gen_mime_type_pattern(strict=False, with_params=False)

MEDIA_TYPE_PATTERN_STRICT = \
    _gen_mime_type_pattern(strict=True, with_params=True)
MEDIA_TYPE_PATTERN_PERMISSIVE = \
    _gen_mime_type_pattern(strict=False, with_params=True)


class Unicode(SimpleModel):
    """The type to represent human-readable data. Its native format is `unicode`
    or `str` with given encoding.
    """

    __type_name__ = 'string'
    Value = six.text_type

    class Attributes(SimpleModel.Attributes):
        """Customizable attributes of the :class:`spyne.model.primitive.Unicode`
        type."""

        min_len = 0
        """Minimum length of string. Can be set to any positive integer"""

        max_len = decimal.Decimal('inf')
        """Maximum length of string. Can be set to ``decimal.Decimal('inf')`` to
        accept strings of arbitrary length. You may also need to adjust
        :const:`spyne.server.wsgi.MAX_CONTENT_LENGTH`."""

        pattern = None
        """A regular expression that matches the whole string. See here for more
        info: http://www.regular-expressions.info/xml.html"""

        unicode_pattern = None
        """Same as ``pattern``, but, will be compiled with ``re.UNICODE``.
        See: https://docs.python.org/2/library/re.html#re.UNICODE"""

        encoding = None
        """The encoding of binary data this class may have to deal with."""

        unicode_errors = 'strict'
        """The argument to the ``unicode`` builtin; one of 'strict', 'replace'
        or 'ignore'."""

        format = None
        """A regular python string formatting string. See here:
        http://docs.python.org/library/stdtypes.html#string-formatting"""

        cast = None
        """Type override callable for casting non-unicode input to unicode."""

    def __new__(cls, *args, **kwargs):
        assert len(args) <= 1

        if len(args) == 1:
            kwargs['max_len'] = args[0]

        retval = SimpleModel.__new__(cls,  ** kwargs)

        return retval

    @staticmethod
    def is_default(cls):
        return (    SimpleModel.is_default(cls)
                and cls.Attributes.min_len == Unicode.Attributes.min_len
                and cls.Attributes.max_len == Unicode.Attributes.max_len
                and cls.Attributes.pattern == Unicode.Attributes.pattern
            )

    @staticmethod
    def validate_string(cls, value):
        return (     SimpleModel.validate_string(cls, value)
            and (value is None or (
                cls.Attributes.min_len <= len(value) <= cls.Attributes.max_len
            )))

    @staticmethod
    def validate_native(cls, value):
        return (SimpleModel.validate_native(cls, value)
            and (value is None or (
                re_match_with_span(cls.Attributes, value)
            )))


class String(Unicode):
    pass

if not six.PY2:
    String = Unicode


class AnyUri(Unicode):
    """A special kind of String type designed to hold an uri."""

    __type_name__ = 'anyURI'

    class Attributes(String.Attributes):
        text = None
        """The text shown in link."""

        anchor_class = None
        """The class of the generated <a> tag."""

    class Value(object):
        """A special object that is just a better way of carrying the
        information carried with a link.

        :param href: The uri string.
        :param text: The text data that goes with the link. This is a
            ``str`` or a ``unicode`` instance.
        :param content: The structured data that goes with the link. This is an
            `lxml.etree.Element` instance.
        """

        def __init__(self, href, text=None, content=None):
            self.href = href
            self.text = text
            self.content = content

        def __repr__(self):
            return "Uri(href={0!r}, text={1!r}, content={2!r})" \
                                     .format(self.href, self.text, self.content)


class ImageUri(AnyUri):
    """A special kind of String that holds the uri of an image."""


def _uuid_validate_string(cls, value):
    return (     SimpleModel.validate_string(cls, value)
        and (value is None or (
            cls.Attributes.min_len <= len(value) <= cls.Attributes.max_len
            and re_match_with_span(cls.Attributes, value)
        )))


def _Tuuid_validate(key):
    from uuid import UUID

    def _uvalid(cls, v):
        try:
            UUID(**{key:v})
        except ValueError:
            return False
        return True
    return _uvalid


_uuid_validate = {
    None: _uuid_validate_string,
    'hex': _Tuuid_validate('hex'),
    'urn': _Tuuid_validate('urn'),
    six.binary_type: _Tuuid_validate('bytes'),
    'bytes': _Tuuid_validate('bytes'),
    'bytes_le': _Tuuid_validate('bytes_le'),
    'fields': _Tuuid_validate('fields'),
    int: _Tuuid_validate('int'),
    'int': _Tuuid_validate('int'),
}


class Uuid(Unicode(pattern=UUID_PATTERN)):
    """Unicode subclass for Universially-Unique Identifiers."""

    __namespace__ = 'http://spyne.io/schema'
    __type_name__ = 'uuid'
    Value = uuid.UUID

    class Attributes(Unicode(pattern=UUID_PATTERN).Attributes):
        serialize_as = None

    @staticmethod
    def validate_string(cls, value):
        return _uuid_validate[cls.Attributes.serialize_as](cls, value)

    @staticmethod
    def validate_native(cls, value):
        return SimpleModel.validate_native(cls, value)


class Ltree(Unicode(LTREE_OPTIMAL_SIZE, unicode_pattern=LTREE_PATTERN)):
    """A special kind of String type designed to hold the Ltree type from
    Postgresql."""

    __namespace__ = 'http://spyne.io/schema'
    __type_name__ = 'ltreeString'


class LtreeLarge(Unicode(LTREE_MAXIMUM_SIZE, unicode_pattern=LTREE_PATTERN)):
    """A special kind of String type designed to hold the Ltree type from
    Postgresql."""

    __namespace__ = 'http://spyne.io/schema'
    __type_name__ = 'largeLtreeString'


class MimeTypeStrict(Unicode(unicode_pattern=MIME_TYPE_PATTERN_STRICT)):
    """A special kind of String type designed to hold a mime type as defined
    by IANA."""

    __namespace__ = 'http://spyne.io/schema'
    __type_name__ = 'strictMimeTypeString'


class MimeType(Unicode(unicode_pattern=MIME_TYPE_PATTERN_PERMISSIVE)):
    """A special kind of String type designed to hold a forward-compatible
    mime type that can have any string as main type."""

    __namespace__ = 'http://spyne.io/schema'
    __type_name__ = 'mimeTypeString'


class MediaTypeStrict(Unicode(unicode_pattern=MEDIA_TYPE_PATTERN_STRICT)):
    """A special kind of String type designed to hold a mime type as defined
    by IANA followed by arbitrary parameters.

    See: https://tools.ietf.org/html/rfc7231#section-3.1.1.1"""

    __namespace__ = 'http://spyne.io/schema'
    __type_name__ = 'strictMediaTypeString'


class MediaType(Unicode(unicode_pattern=MEDIA_TYPE_PATTERN_PERMISSIVE)):
    """A special kind of String type designed to hold a forward-compatible
    media type that can have any string as main type. A media type is
    essentially a mime type plus parameters.

    See: https://tools.ietf.org/html/rfc7231#section-3.1.1.1"""

    __namespace__ = 'http://spyne.io/schema'
    __type_name__ = 'mediaTypeString'


if not six.PY2:
    NATIVE_MAP.update({
        str: Unicode,
    })

else:
    NATIVE_MAP.update({
        str: String,
        unicode: Unicode,
    })




#
# spyne - Copyright (C) Spyne contributors.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
#

"""This module contains the ModelBase class and other building blocks for
defining models.
"""

from __future__ import print_function

import logging
logger = logging.getLogger(__name__)

import re
import decimal
import threading

import spyne.const.xml

from copy import deepcopy
from collections import OrderedDict

from spyne import const
from spyne.util import Break, six
from spyne.util.cdict import cdict
from spyne.util.odict import odict

from spyne.const.xml import DEFAULT_NS


class Ignored(object):
    """When returned as a real rpc response, this is equivalent to returning
    None. However, direct method invocations (and NullServer) get the return
    value. It can be used for tests and from hooks."""

    __slots__ = ('args', 'kwargs')

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def __eq__(self, other):
        return      isinstance(other, Ignored) \
                    and self.args == other.args and self.kwargs == other.kwargs

    def __ne__(self, other):
        return not (isinstance(other, Ignored) \
                    and self.args == other.args and self.kwargs == other.kwargs)


def _decode_pa_dict(d):
    """Decodes dict passed to prot_attrs.

    >>> _decode_pa_dict({})
    cdict({})
    >>> _decode_pa_dict({1: 2)})
    cdict({1: 2})
    >>> _decode_pa_dict({(1,2): 3)})
    cdict({1: 3, 2: 3})
    """

    retval = cdict()
    for k, v in d.items():
        if isinstance(k, (frozenset, tuple)):
            for subk in k:
                retval[subk] = v

    for k, v in d.items():
        if not isinstance(k, (frozenset, tuple)):
            retval[k] = v

    return retval


class AttributesMeta(type(object)):
    NULLABLE_DEFAULT = True

    def __new__(cls, cls_name, cls_bases, cls_dict):
        # Mapper args should not be inherited.
        if not 'sqla_mapper_args' in cls_dict:
            cls_dict['sqla_mapper_args'] = None

        rd = {}
        for k in list(cls_dict.keys()):
            if k in ('parser', 'cast'):
                rd['parser'] = cls_dict.pop(k)
                continue

            if k in ('sanitize', 'sanitizer'):
                rd['sanitizer'] = cls_dict.pop(k)
                continue

            if k == 'logged':
                rd['logged'] = cls_dict.pop(k)
                continue

        retval = super(AttributesMeta, cls).__new__(cls, cls_name, cls_bases,
                                                                       cls_dict)

        for k, v in rd.items():
            if v is None:
                setattr(retval, k, None)
            else:
                setattr(retval, k, staticmethod(v))

        return retval

    def __init__(self, cls_name, cls_bases, cls_dict):
        # you will probably want to look at ModelBase._s_customize as well.
        if not hasattr(self, '_method_config_do'):
            self._method_config_do = None

        nullable = cls_dict.get('nullable', None)
        nillable = cls_dict.get('nillable', None)
        if nullable is not None:
            assert nillable is None or nullable == nillable
            self._nullable = nullable

        elif nillable is not None:
            assert nullable is None or nullable == nillable
            self._nullable = nillable

        if not hasattr(self, '_nullable'):
            self._nullable = None

        if not hasattr(self, '_default_factory'):
            self._default_factory = None

        if not hasattr(self, '_html_cloth'):
            self._html_cloth = None
        if not hasattr(self, '_html_root_cloth'):
            self._html_root_cloth = None

        if 'html_cloth' in cls_dict:
            self.set_html_cloth(cls_dict.pop('html_cloth'))
        if 'html_root_cloth' in cls_dict:
            self.set_html_cloth(cls_dict.pop('html_root_cloth'))

        if not hasattr(self, '_xml_cloth'):
            self._xml_cloth = None
        if not hasattr(self, '_xml_root_cloth'):
            self._xml_root_cloth = None

        if 'xml_cloth' in cls_dict:
            self.set_xml_cloth(cls_dict.pop('xml_cloth'))

        if 'xml_root_cloth' in cls_dict:
            self.set_xml_cloth(cls_dict.pop('xml_root_cloth'))

        if 'method_config_do' in cls_dict and \
                                       cls_dict['method_config_do'] is not None:
            cls_dict['method_config_do'] = \
                                      staticmethod(cls_dict['method_config_do'])

        super(AttributesMeta, self).__init__(cls_name, cls_bases, cls_dict)

    def get_nullable(self):
        return (self._nullable if self._nullable is not None else
                                                          self.NULLABLE_DEFAULT)

    def set_nullable(self, what):
        self._nullable = what

    nullable = property(get_nullable, set_nullable)

    def get_nillable(self):
        return self.nullable

    def set_nillable(self, what):
        self.nullable = what

    nillable = property(get_nillable, set_nillable)

    def get_default_factory(self):
        return self._default_factory

    def set_default_factory(self, what):
        self._default_factory = staticmethod(what)

    default_factory = property(get_default_factory, set_default_factory)

    def get_html_cloth(self):
        return self._html_cloth
    def set_html_cloth(self, what):
        from spyne.protocol.cloth.to_cloth import ClothParserMixin
        cm = ClothParserMixin.from_html_cloth(what)
        if cm._root_cloth is not None:
            self._html_root_cloth = cm._root_cloth
        elif cm._cloth is not None:
            self._html_cloth = cm._cloth
        else:
            raise Exception("%r is not a suitable cloth", what)
    html_cloth = property(get_html_cloth, set_html_cloth)

    def get_html_root_cloth(self):
        return self._html_root_cloth
    html_root_cloth = property(get_html_root_cloth)

    def get_xml_cloth(self):
        return self._xml_cloth
    def set_xml_cloth(self, what):
        from spyne.protocol.cloth.to_cloth import ClothParserMixin
        cm = ClothParserMixin.from_xml_cloth(what)
        if cm._root_cloth is not None:
            self._xml_root_cloth = cm._root_cloth
        elif cm._cloth is not None:
            self._xml_cloth = cm._cloth
        else:
            raise Exception("%r is not a suitable cloth", what)
    xml_cloth = property(get_xml_cloth, set_xml_cloth)

    def get_xml_root_cloth(self):
        return self._xml_root_cloth
    xml_root_cloth = property(get_xml_root_cloth)

    def get_method_config_do(self):
        return self._method_config_do
    def set_method_config_do(self, what):
        if what is None:
            self._method_config_do = None
        else:
            self._method_config_do = staticmethod(what)
    method_config_do = property(get_method_config_do, set_method_config_do)


class ModelBaseMeta(type(object)):
    def __getitem__(self, item):
        return self.customize(**item)

    def customize(self, **kwargs):
        """Duplicates cls and overwrites the values in ``cls.Attributes`` with
        ``**kwargs`` and returns the new class."""

        cls_name, cls_bases, cls_dict = self._s_customize(**kwargs)

        return type(cls_name, cls_bases, cls_dict)


@six.add_metaclass(ModelBaseMeta)
class ModelBase(object):
    """The base class for type markers. It defines the model interface for the
    interface generators to use and also manages class customizations that are
    mainly used for defining constraints on input values.
    """

    __orig__ = None
    """This holds the original class the class .customize()d from. Ie if this is
    None, the class is not a customize()d one."""

    __extends__ = None
    """This holds the original class the class inherited or .customize()d from.
    This is different from __orig__ because it's only set when
    ``cls.is_default(cls) == False``"""

    __namespace__ = None
    """The public namespace of this class. Use ``get_namespace()`` instead of
    accessing it directly."""

    __type_name__ = None
    """The public type name of the class. Use ``get_type_name()`` instead of
    accessing it directly."""

    Value = type(None)
    """The value of this type is an instance of this class"""

    # These are not the xml schema defaults. The xml schema defaults are
    # considered in XmlSchema's add() method. the defaults here are to reflect
    # what people seem to want most.
    #
    # Please note that min_occurs and max_occurs must be validated in the
    # ComplexModelBase deserializer.
    @six.add_metaclass(AttributesMeta)
    class Attributes(object):
        """The class that holds the constraints for the given type."""

        _wrapper = False
        # when skip_wrappers=True is passed to a protocol, these objects
        # are skipped. just for internal use.

        _explicit_type_name = False
        # set to true when type_name is passed to customize() call.

        out_type = None
        """Override serialization type. Usually, this designates the return type
        of the callable in the `sanitizer` attribute. If this is a two-way type,
        it may be a good idea to also use the `parser` attribute to perform
        reverse conversion."""

        default = None
        """The default value if the input is None.

        Please note that this default is UNCONDITIONALLY applied in class
        initializer. It's recommended to at least make an effort to use this
        only in customized classes and not in original models.
        """

        default_factory = None
        """The callable that produces a default value if the value is None.

        The warnings in ``default`` apply here as well."""

        db_default = None
        """The default value used only when persisting the value if it is
        ``None``.

        Only works for primitives. Unlike ``default`` this can also be set to a
        callable that takes no arguments according to SQLAlchemy docs."""

        nillable = None
        """Set this to false to reject null values. Synonyms with
        ``nullable``. True by default. The default value can be changed by
         setting ``AttributesMeta.NULLABLE_DEFAULT``."""

        min_occurs = 0
        """Set this to 1 to make this object mandatory. Can be set to any
        positive integer. Note that an object can still be null or empty, even
        if it's there."""

        max_occurs = 1
        """Can be set to any strictly positive integer. Values greater than 1
        will imply an iterable of objects as native python type. Can be set to
        ``decimal.Decimal("inf")`` for arbitrary number of arguments."""

        schema_tag = spyne.const.xml.XSD('element')
        """The tag used to add a primitives as child to a complex type in the
        xml schema."""

        translations = None
        """A dict that contains locale codes as keys and translations of field
        names to that language as values.
        """

        sub_ns = None
        """An Xml-specific attribute that specifies which namespace should be
        used for field names in classes.
        """

        sub_name = None
        """This specifies which string should be used as field name when this
        type is seriazed under a ComplexModel.
        """

        wsdl_part_name = None
        """This specifies which string should be used as wsdl message part name when this
            type is serialized under a ComplexModel ie."parameters".
        """

        sqla_column_args = None
        """A dict that will be passed to SQLAlchemy's ``Column`` constructor as
        ``**kwargs``.
        """

        exc_mapper = False
        """If true, this field will be excluded from the table mapper of the
        parent class.
        """

        exc_table = False
        """DEPRECATED !!! Use ``exc_db`` instead."""

        exc_db = False
        """If ``True``, this field will not be persisted to the database. This
        attribute only makes sense in a subfield of a ``ComplexModel`` subclass.
        """

        exc_interface = False
        """If `True`, this field will be excluded from the interface
        document."""

        exc = False
        """If `True`, this field will be excluded from all serialization or
         deserialization operations. See `prot_attrs` to make this only apply to
         a specific protocol class or instance."""

        logged = True
        """If `False`, this object will be ignored in ``log_repr``, mostly used
        for logging purposes.

        * Primitives can have logger=``'...'`` which will
        always log the value as ``(...)``.

        * ``AnyDict`` can have one of
        ``('keys', 'keys-full', 'values', 'values-full, 'full')`` as logger
        value where for ``'keys'`` and ``'values'`` the output of ``keys()``
        and ``values()`` will be logged up to MAX_DICT_ELEMENT_NUM number of
        elements and for ``'full'`` variants, all of the contents of the dict
        will be logged will be logged

        * ``Array`` can also have ``logger='full'`` where all of the value
        will be logged where as for simple ``logger=True`` only
        MAX_ARRAY_ELEMENT_NUM elements will be logged.

        * For ``ComplexModel`` subclasses sent as first value to log_repr,
        ``logger=False`` means a string of form ``ClassName(...)`` will  be
        logged.
        """

        sanitizer = None
        """A callable that takes the associated native type and returns the
        parsed value. Only called during serialization."""

        parser = None
        """A callable that takes the associated native type and returns the
        parsed value. Only called during deserialization."""

        unique = None
        """If True, this object will be set as unique in the database schema
        with default indexing options. If the value is a string, it will be
        used as the indexing method to create the unique index. See sqlalchemy
        documentation on how to create multi-column unique constraints.
        """

        db_type = None
        """When not None, it overrides Spyne's own mapping from Spyne types to
        SQLAlchemy types. It's a standard SQLAlchemy type marker, e.g.
        ``sqlalchemy.Integer``.
        """

        table_name = None
        """Database table name."""

        xml_choice_group = None
        """When not None, shares the same <choice> tag with fields with the same
        xml_choice_group value.
        """

        index = None
        """Can be ``True``, a string, or a tuple of two strings.

        * If True, this object will be set as indexed in the database schema
          with default options.

        * If the value is a string, the value will denote the indexing method
          used by the database. Should be one of:

            ('btree', 'gin', 'gist', 'hash', 'spgist')

          See: http://www.postgresql.org/docs/9.2/static/indexes-types.html

        * If the value is a tuple of two strings, the first value will denote
          the index name and the second value will denote the indexing method as
          above.
        """

        read_only = False
        """If True, the attribute won't be initialized from outside values.
        Set this to ``True`` for e.g. read-only properties."""

        prot_attrs = None
        """Customize child attributes for protocols. It's a dict of dicts.
        The key is either a ProtocolBase subclass or a ProtocolBase instance.
        Instances override classes."""

        pa = None
        """Alias for prot_attrs."""

        empty_is_none = False
        """When the incoming object is empty (e.g. '' for strings) treat it as
        None. No effect (yet) for outgoing values."""

        order = None
        """An integer that's passed to ``_type_info.insert()`` as first argument
         when not None. ``.append()`` is used otherwise."""

        validate_on_assignment = False
        """Perform validation on assignment (i.e. all the time) instead of on
        just serialization"""

        polymap = {}
        """A dict of classes that override polymorphic substitions for classes
        given as keys to classes given as values."""


    class Annotations(object):
        """The class that holds the annotations for the given type."""

        __use_parent_doc__ = False
        """If equal to True and doc is empty, Annotations will use __doc__
        from parent. Set it to False to avoid this mechanism. This is a
        convenience option"""

        doc = ""
        """The public documentation for the given type."""

        appinfo = None
        """Any object that carries app-specific info."""

    class Empty(object):
        pass

    _force_own_namespace = None

    @classmethod
    def ancestors(cls):
        """Returns a list of parent classes in child-to-parent order."""

        retval = []

        extends = cls.__extends__
        while extends is not None:
            retval.append(extends)
            extends = extends.__extends__

        return retval

    @staticmethod
    def is_default(cls):
        return True

    @classmethod
    def get_namespace_prefix(cls, interface):
        """Returns the namespace prefix for the given interface. The
        get_namespace_prefix of the interface class generates a prefix if none
        is defined.
        """

        ns = cls.get_namespace()

        retval = interface.get_namespace_prefix(ns)

        return retval

    @classmethod
    def get_namespace(cls):
        """Returns the namespace of the class. Defaults to the python module
        name."""

        return cls.__namespace__

    @classmethod
    def _fill_empty_type_name(cls, parent_ns, parent_tn, k):
        cls.__namespace__ = parent_ns

        cls.__type_name__ = "%s_%s%s" % (parent_tn, k, const.TYPE_SUFFIX)
        extends = cls.__extends__
        while extends is not None and extends.__type_name__ is ModelBase.Empty:
            cls.__extends__._fill_empty_type_name(cls.get_namespace(),
                               cls.get_type_name(), k + const.PARENT_SUFFIX)
            extends = extends.__extends__

    # TODO: rename to "resolve_identifier"
    @staticmethod
    def resolve_namespace(cls, default_ns, tags=None):
        """This call finalizes the namespace assignment. The default namespace
        is not available until the application calls populate_interface method
        of the interface generator.
        """

        if tags is None:
            tags = set()
        elif cls in tags:
            return False
        tags.add(cls)

        if cls.__namespace__ is spyne.const.xml.DEFAULT_NS:
            cls.__namespace__ = default_ns

        if (cls.__namespace__ in spyne.const.xml.PREFMAP and
                                                       not cls.is_default(cls)):
            cls.__namespace__ = default_ns

        if cls.__namespace__ is None:
            ret = []
            for f in cls.__module__.split('.'):
                if f.startswith('_'):
                    break
                else:
                    ret.append(f)

            cls.__namespace__ = '.'.join(ret)

        if cls.__namespace__ is None or len(cls.__namespace__) == 0:
            cls.__namespace__ = default_ns

        if cls.__namespace__ is None or len(cls.__namespace__) == 0:
            raise ValueError("You need to explicitly set %r.__namespace__" % cls)

        # print("    resolve ns for %r to %r" % (cls, cls.__namespace__))

        if getattr(cls, '__extends__', None) != None:
            cls.__extends__.resolve_namespace(cls.__extends__, default_ns, tags)

        return True

    @classmethod
    def get_type_name(cls):
        """Returns the class name unless the __type_name__ attribute is defined.
        """

        retval = cls.__type_name__
        if retval is None:
            retval = cls.__name__

        return retval

    # FIXME: Rename this to get_type_name_with_ns_pref
    @classmethod
    def get_type_name_ns(cls, interface):
        """Returns the type name with a namespace prefix, separated by a column.
        """

        if cls.get_namespace() != None:
            return "%s:%s" % (cls.get_namespace_prefix(interface),
                                                            cls.get_type_name())

    @classmethod
    def get_element_name(cls):
        return cls.Attributes.sub_name or cls.get_type_name()

    @classmethod
    def get_wsdl_part_name(cls):
        return cls.Attributes.wsdl_part_name or cls.get_element_name()

    @classmethod
    def get_element_name_ns(cls, interface):
        ns = cls.Attributes.sub_ns or cls.get_namespace()
        if ns is DEFAULT_NS:
            ns = interface.get_tns()
        if ns is not None:
            pref = interface.get_namespace_prefix(ns)
            return "%s:%s" % (pref, cls.get_element_name())

    @classmethod
    def to_bytes(cls, value):
        """
        Returns str(value). This should be overridden if this is not enough.
        """
        return six.binary_type(value)

    @classmethod
    def to_unicode(cls, value):
        """
        Returns unicode(value). This should be overridden if this is not enough.
        """
        return six.text_type(value)

    @classmethod
    def get_documentation(cls):
        if cls.Annotations.doc:
            return cls.Annotations.doc
        elif cls.Annotations.__use_parent_doc__:
            return cls.__doc__
        else:
            return ''

    @classmethod
    def _s_customize(cls, **kwargs):
        """Sanitizes customization parameters of the class it belongs to.
        Doesn't perform any actual customization.
        """

        def _log_debug(s, *args):
            logger.debug("\t%s: %s" % (cls.get_type_name(), s), *args)

        cls_dict = odict({'__module__': cls.__module__, '__doc__': cls.__doc__})

        if getattr(cls, '__orig__', None) is None:
            cls_dict['__orig__'] = cls
        else:
            cls_dict['__orig__'] = cls.__orig__

        class Attributes(cls.Attributes):
            _explicit_type_name = False

        if cls.Attributes.translations is None:
            Attributes.translations = {}

        if cls.Attributes.sqla_column_args is None:
            Attributes.sqla_column_args = (), {}
        else:
            Attributes.sqla_column_args = deepcopy(
                                                cls.Attributes.sqla_column_args)

        cls_dict['Attributes'] = Attributes

        # properties get reset every time a new class is defined. So we need
        # to reinitialize them explicitly.
        for k in ('nillable', '_xml_cloth', '_xml_root_cloth', '_html_cloth',
                                                            '_html_root_cloth'):
            v = getattr(cls.Attributes, k)
            if v is not None:
                setattr(Attributes, k, v)

        class Annotations(cls.Annotations):
            pass
        cls_dict['Annotations'] = Annotations

        # get protocol attrs
        prot = kwargs.get('protocol', None)
        if prot is None:
            prot = kwargs.get('prot', None)

        if prot is None:
            prot = kwargs.get('p', None)

        if prot is not None and len(prot.type_attrs) > 0:
            # if there is a class customization from protocol, do it

            type_attrs = prot.type_attrs.copy()
            type_attrs.update(kwargs)
            _log_debug("kwargs %r => %r from prot typeattr %r",
                                            kwargs, type_attrs, prot.type_attrs)
            kwargs = type_attrs

        # the ones that wrap values in staticmethod() should be added to
        # AttributesMeta initializer
        for k, v in kwargs.items():
            if k.startswith('_'):
                _log_debug("ignoring '%s' because of leading underscore", k)
                continue

            if k in ('protocol', 'prot', 'p'):
                Attributes.prot = v
                _log_debug("setting prot=%r", v)

            elif k in ('voa', 'validate_on_assignment'):
                Attributes.validate_on_assignment = v
                _log_debug("setting voa=%r", v)

            elif k in ('parser', 'in_cast'):
                setattr(Attributes, 'parser', staticmethod(v))
                _log_debug("setting %s=%r", k, v)

            elif k in ('sanitize', 'sanitizer', 'out_cast'):
                setattr(Attributes, 'sanitizer', staticmethod(v))
                _log_debug("setting %s=%r as sanitizer", k, v)

            elif k == 'logged':
                setattr(Attributes, 'logged', staticmethod(v))
                _log_debug("setting %s=%r as log sanitizer", k, v)

            elif k in ("doc", "appinfo"):
                setattr(Annotations, k, v)
                _log_debug("setting Annotations.%s=%r", k, v)

            elif k in ('primary_key', 'pk'):
                setattr(Attributes, 'primary_key', v)
                Attributes.sqla_column_args[-1]['primary_key'] = v
                _log_debug("setting primary_key=%r", v)

            elif k in ('protocol_attrs', 'prot_attrs', 'pa'):
                setattr(Attributes, 'prot_attrs', _decode_pa_dict(v))
                _log_debug("setting prot_attrs=%r", v)

            elif k in ('foreign_key', 'fk'):
                from sqlalchemy.schema import ForeignKey
                t, d = Attributes.sqla_column_args
                fkt = (ForeignKey(v),)
                new_v = (t + fkt, d)
                Attributes.sqla_column_args = new_v
                _log_debug("setting sqla_column_args=%r", new_v)

            elif k in ('autoincrement', 'onupdate', 'server_default'):
                Attributes.sqla_column_args[-1][k] = v
                _log_debug("adding %s=%r to Attributes.sqla_column_args", k, v)

            elif k == 'values_dict':
                assert not 'values' in v, "`values` and `values_dict` can't be" \
                                          "specified at the same time"

                if not isinstance(v, dict):
                    # our odict has one nasty implicit behaviour: setitem on
                    # int keys is treated as array indexes, not dict keys. so
                    # dicts with int indexes can't work with odict. so we use
                    # the one from stdlib
                    v = OrderedDict(v)

                Attributes.values = list(v.keys())
                Attributes.values_dict = v
                _log_debug("setting values=%r, values_dict=%r",
                                      Attributes.values, Attributes.values_dict)

            elif k == 'exc_table':
                Attributes.exc_table = v
                Attributes.exc_db = v
                _log_debug("setting exc_table=%r, exc_db=%r", v, v)

            elif k == 'max_occurs' and v in ('unbounded', 'inf', float('inf')):
                new_v = decimal.Decimal('inf')
                setattr(Attributes, k, new_v)
                _log_debug("setting max_occurs=%r", new_v)

            elif k == 'type_name':
                Attributes._explicit_type_name = True
                _log_debug("setting _explicit_type_name=True because "
                                                          "we have 'type_name'")

            else:
                setattr(Attributes, k, v)
                _log_debug("setting %s=%r", k, v)

        return (cls.__name__, (cls,), cls_dict)

    @staticmethod
    def validate_string(cls, value):
        """Override this method to do your own input validation on the input
        string. This is called before converting the incoming string to the
        native python value."""

        return (cls.Attributes.nillable or value is not None)

    @staticmethod
    def validate_native(cls, value):
        """Override this method to do your own input validation on the native
        value. This is called after converting the incoming string to the
        native python value."""

        return (cls.Attributes.nullable or value is not None)


class Null(ModelBase):
    pass


class SimpleModelAttributesMeta(AttributesMeta):
    def __init__(self, cls_name, cls_bases, cls_dict):
        super(SimpleModelAttributesMeta, self).__init__(cls_name, cls_bases,
                                                                       cls_dict)
        if getattr(self, '_pattern', None) is None:
            self._pattern = None

    def get_pattern(self):
        return self._pattern

    def set_pattern(self, pattern):
        self._pattern = pattern
        if pattern is not None:
            self._pattern_re = re.compile(pattern)

    pattern = property(get_pattern, set_pattern)

    def get_unicode_pattern(self):
        return self._pattern

    def set_unicode_pattern(self, pattern):
        self._pattern = pattern
        if pattern is not None:
            self._pattern_re = re.compile(pattern, re.UNICODE)

    unicode_pattern = property(get_unicode_pattern, set_unicode_pattern)
    upattern = property(get_unicode_pattern, set_unicode_pattern)


class SimpleModel(ModelBase):
    """The base class for primitives."""

    __namespace__ = "http://www.w3.org/2001/XMLSchema"

    @six.add_metaclass(SimpleModelAttributesMeta)
    class Attributes(ModelBase.Attributes):
        """The class that holds the constraints for the given type."""

        values = set()
        """The set of possible values for this type."""

        # some hacks are done in _s_customize to make `values_dict`
        # behave like `values`
        values_dict = dict()
        """The dict of possible values for this type. Dict keys are values and
        dict values are either a single string or a translation dict."""

        _pattern_re = None

    def __new__(cls, **kwargs):
        """Overriden so that any attempt to instantiate a primitive will return
        a customized class instead of an instance.

        See spyne.model.base.ModelBase for more information.
        """

        return cls.customize(**kwargs)

    @classmethod
    def customize(cls, **kwargs):
        """Duplicates cls and overwrites the values in ``cls.Attributes`` with
        ``**kwargs`` and returns the new class."""

        cls_name, cls_bases, cls_dict = cls._s_customize(**kwargs)

        retval = type(cls_name, cls_bases, cls_dict)

        if not retval.is_default(retval):
            retval.__extends__ = cls
            retval.__type_name__ = kwargs.get("type_name", ModelBase.Empty)
            if 'type_name' in kwargs:
                logger.debug("Type name for %r was overridden as '%s'",
                                                   retval, retval.__type_name__)

        retval.resolve_namespace(retval, kwargs.get('__namespace__'))

        return retval

    @staticmethod
    def is_default(cls):
        return (cls.Attributes.values == SimpleModel.Attributes.values)

    @staticmethod
    def validate_native(cls, value):
        return (ModelBase.validate_native(cls, value)
                and (
                    cls.Attributes.values is None or
                    len(cls.Attributes.values) == 0 or (
                        (value is None     and cls.Attributes.nillable) or
                        (value is not None and value in cls.Attributes.values)
                    )
                )
            )


class PushBase(object):
    def __init__(self, callback=None, errback=None):
        self.orig_thread = threading.current_thread()

        self._cb = callback
        self._eb = errback

        self.length = 0
        self.ctx = None
        self.app = None
        self.gen = None
        self._cb_finish = None
        self._eb_finish = None
        self.interim = False

    def _init(self, ctx, gen, _cb_finish, _eb_finish, interim):
        self.length = 0

        self.ctx = ctx
        self.app = ctx.app

        self.gen = gen

        self._cb_finish = _cb_finish
        self._eb_finish = _eb_finish

        self.interim = interim

    def init(self, ctx, gen, _cb_finish, _eb_finish, interim):
        self._init(ctx, gen, _cb_finish, _eb_finish, interim)
        if self._cb is not None:
            return self._cb(self)

    def __len__(self):
        return self.length

    def append(self, inst):
        self.gen.send(inst)
        self.length += 1

    def extend(self, insts):
        for inst in insts:
            self.gen.send(inst)
            self.length += 1

    def close(self):
        try:
            self.gen.throw(Break())
        except (Break, StopIteration, GeneratorExit):
            pass
        self._cb_finish()


class xml:
    """Compound option object for xml serialization. It's meant to be passed to
    :func:`ComplexModelBase.Attributes.store_as`.

    :param root_tag: Root tag of the xml element that contains the field values.
    :param no_ns: When true, the xml document is stripped from namespace
        information. This is generally a stupid thing to do. Use with caution.
    """

    def __init__(self, root_tag=None, no_ns=False, pretty_print=False):
        self.root_tag = root_tag
        self.no_ns = no_ns
        self.pretty_print = pretty_print


class table:
    """Compound option object for for storing the class instance as in row in a
    table in a relational database. It's meant to be passed to
    :func:`ComplexModelBase.Attributes.store_as`.

    :param multi: When False, configures a one-to-many relationship where the
        child table has a foreign key to the parent. When not ``False``,
        configures a many-to-many relationship by creating an intermediate
        relation table that has foreign keys to both parent and child classes
        and generates a table name automatically. When ``True``, the table name
        is generated automatically. Otherwise, it should be a string, as the
        value is used as the name of the intermediate table.
    :param left: Name of the left join column.
    :param right: Name of the right join column.
    :param backref: See https://docs.sqlalchemy.org/en/13/orm/relationship_api.html?highlight=lazy#sqlalchemy.orm.relationship.params.backref
    :param cascade: https://docs.sqlalchemy.org/en/13/orm/relationship_api.html?highlight=lazy#sqlalchemy.orm.relationship.params.cascade
    :param lazy: See https://docs.sqlalchemy.org/en/13/orm/relationship_api.html?highlight=lazy#sqlalchemy.orm.relationship.params.lazy
    :param back_populates: See https://docs.sqlalchemy.org/en/13/orm/relationship_api.html?highlight=lazy#sqlalchemy.orm.relationship.params.back_populates
    """

    def __init__(self, multi=False, left=None, right=None, backref=None,
            id_backref=None, cascade=False, lazy='select', back_populates=None,
                       fk_left_deferrable=None, fk_left_initially=None,
                       fk_right_deferrable=None, fk_right_initially=None,
                       fk_left_ondelete=None, fk_left_onupdate=None,
                       fk_right_ondelete=None, fk_right_onupdate=None,
                       explicit_join=False, order_by=False, single_parent=None):
        self.multi = multi
        self.left = left
        self.right = right
        self.backref = backref
        self.id_backref = id_backref
        self.cascade = cascade
        self.lazy = lazy
        self.back_populates = back_populates
        self.fk_left_deferrable = fk_left_deferrable
        self.fk_left_initially = fk_left_initially
        self.fk_right_deferrable = fk_right_deferrable
        self.fk_right_initially = fk_right_initially
        self.fk_left_ondelete = fk_left_ondelete
        self.fk_left_onupdate = fk_left_onupdate
        self.fk_right_ondelete = fk_right_ondelete
        self.fk_right_onupdate = fk_right_onupdate
        self.explicit_join = explicit_join
        self.order_by = order_by
        self.single_parent = single_parent


class json:
    """Compound option object for json serialization. It's meant to be passed to
    :func:`ComplexModelBase.Attributes.store_as`.

    Make sure you don't mix this with the json package when importing.
    """

    def __init__(self, ignore_wrappers=True, complex_as=dict):
        if ignore_wrappers != True:
            raise NotImplementedError("ignore_wrappers != True")
        self.ignore_wrappers = ignore_wrappers
        self.complex_as = complex_as


class jsonb:
    """Compound option object for jsonb serialization. It's meant to be passed
    to :func:`ComplexModelBase.Attributes.store_as`.
    """

    def __init__(self, ignore_wrappers=True, complex_as=dict):
        if ignore_wrappers != True:
            raise NotImplementedError("ignore_wrappers != True")
        self.ignore_wrappers = ignore_wrappers
        self.complex_as = complex_as


class msgpack:
    """Compound option object for msgpack serialization. It's meant to be passed
    to :func:`ComplexModelBase.Attributes.store_as`.

    Make sure you don't mix this with the msgpack package when importing.
    """
    def __init__(self):
        pass


PSSM_VALUES = {'json': json, 'jsonb': jsonb, 'xml': xml,
                                             'msgpack': msgpack, 'table': table}


def apply_pssm(val):
    if val is not None:
        val_c = PSSM_VALUES.get(val, None)
        if val_c is None:
            assert isinstance(val, tuple(PSSM_VALUES.values())), \
             "'store_as' should be one of: %r or an instance of %r not %r" \
             % (tuple(PSSM_VALUES.keys()), tuple(PSSM_VALUES.values()), val)

            return val
        return val_c()
